import glob
import os
import time
from tempfile import NamedTemporaryFile, TemporaryDirectory

import numpy as np

try:
    import vedo.vtkclasses as vtk
except ImportError:
    import vtkmodules.all as vtk

import vedo
from vedo import settings
from vedo import colors
from vedo import utils
from vedo.assembly import Assembly
from vedo.picture import Picture
from vedo.pointcloud import Points
from vedo.mesh import Mesh
from vedo.volume import Volume

__docformat__ = "google"

__doc__ = """
Submodule to read/write meshes and other objects in different formats,
and other I/O functionalities.
"""

__all__ = [
    "load",
    "download",
    "gunzip",
    "loadStructuredPoints",
    "loadStructuredGrid",
    "loadRectilinearGrid",
    "loadUnStructuredGrid",
    "load_transform",
    "write_transform",
    "write",
    "export_window",
    "import_window",
    "screenshot",
    "ask",
    "Video",
]


# example web page for X3D
_x3d_html = """
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title> vedo with x3d </title>

  <!-- THESE ARE THE RELEVANT LINES: -->
  <script src='https://www.x3dom.org/download/x3dom.js'> </script>
  <link rel='stylesheet' type='text/css' href='https://www.x3dom.org/download/x3dom.css'/>

  <style>
     table, td, th { border: 1px solid black; background-color: powderblue;}
     table {width: 70%; border-collapse: collapse;}
     table th {width: 35%;}
  </style>
</head>

<body style="font-family: Verdana">
  <h1>Example html generated by vedo</h1>
  This example loads a 3D scene from file ~fileoutput generated by
  <a href="https://github.com/marcomusy/vedo">vedo</a>
  (see <a href="https://github.com/marcomusy/vedo/tree/master/examples/other/export_x3d.py">export_x3d.py</a>).
  <br><br>


  <!-- THESE ARE THE RELEVANT LINES: -->
  <x3d width='~widthpx' height='~heightpx'>
     <scene>
        <Inline url="~fileoutput"> </Inline>
     </scene>
  </x3d>

  <h3>Nothing shows up above this line?</h3>
  Enable your browser to load local files:
  <br><b>Firefox</b>: type <code>about:config</code> in the URL bar and
  change <code>privacy.file_unique_origin</code> from <code>True</code> to <code>False</code>
  <br><b>Chrome</b>: from terminal type:
  <code>google-chrome --enable-webgl --allow-file-access-from-files</code>
  (see <a href="https://cmatskas.com/interacting-with-local-data-files-using-chrome/">here</a>)

  <br>
  <h3>Controls:</h3>
  <h4><strong>Examine Mode (activate with key 'e'):</strong></h4>
  <table>
     <tbody>
        <tr class="even description">
           <th>Button</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>Left Button / Left Button + Shift</td>
           <td>Rotate</td>
        </tr>
        <tr>
           <td>Mid Button / Left Button + Ctl</td>
           <td>Pan</td>
        </tr>
        <tr>
           <td>Right Button / Wheel / Left Button + Alt</td>
           <td>Zoom</td>
        </tr>
        <tr>
           <td>Left double click</td>
           <td>Set center of rotation</td>
        </tr>
     </tbody>
  </table>
  <h4><strong>Walk Mode (activate with key 'w'):</strong></h4>
  <table>
     <tbody>
        <tr class="even description">
           <th>Button</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>Left Button</td>
           <td>Move forward</td>
        </tr>
        <tr>
           <td>Right Button</td>
           <td>Move backward</td>
        </tr>
     </tbody>
  </table>
  <h4><strong>Fly Mode (activate with key 'f'):</strong></h4>
  <table>
     <tbody>
        <tr class="even description">
           <th>Button</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>Left Button</td>
           <td>Move forward</td>
        </tr>
        <tr>
           <td>Right Button</td>
           <td>Move backward</td>
        </tr>
     </tbody>
  </table>
  <h3>Non-interactive camera movement</h3>
  <table>
     <tbody>
        <tr class="even description">
           <th>Key</th>
           <th>Function</th>
        </tr>
        <tr>
           <td>r</td>
           <td>reset view</td>
        </tr>
        <tr>
           <td>a</td>
           <td>show all</td>
        </tr>
        <tr>
           <td>u</td>
           <td>upright</td>
        </tr>
     </tbody>
  </table>
</body>
</html>
"""


def load(inputobj, unpack=True, force=False):
    """
    Load any vedo objects from file or from the web.

    The output will depend on the file extension. See examples below.
    Unzip is made on the fly, if file ends with `.gz`.
    Can load an object directly from a URL address.

    Arguments:
        unpack : bool
            unpack MultiBlockData into a flat list of objects.

        force : bool
            when downloading a file ignore any previous cached downloads and force a new one.

    Example:
        ```python
        from vedo import dataurl, load, show
        # Return a list of 2 meshes
        g = load([dataurl+'250.vtk', dataurl+'270.vtk'])
        show(g)
        # Return a list of meshes by reading all files in a directory
        # (if directory contains DICOM files then a Volume is returned)
        g = load('mydicomdir/')
        show(g)
        ```
    """
    acts = []
    if utils.is_sequence(inputobj):
        flist = inputobj
    elif isinstance(inputobj, str) and inputobj.startswith("https://"):
        flist = [inputobj]
    else:
        # flist = sorted(glob.glob(inputobj))
        flist = utils.humansort(glob.glob(inputobj))

    for fod in flist:

        if fod.startswith("https://"):
            fod = download(fod, force=force, verbose=False)

        if os.path.isfile(fod):  ### it's a file

            if fod.endswith(".gz"):
                fod = gunzip(fod)

            a = _load_file(fod, unpack)
            acts.append(a)

        elif os.path.isdir(fod):  ### it's a directory or DICOM
            flist = os.listdir(fod)
            if ".dcm" in flist[0]:  ### it's DICOM
                reader = vtk.vtkDICOMImageReader()
                reader.SetDirectoryName(fod)
                reader.Update()
                image = reader.GetOutput()
                actor = Volume(image)

                actor.info["PixelSpacing"] = reader.GetPixelSpacing()
                actor.info["Width"] = reader.GetWidth()
                actor.info["Height"] = reader.GetHeight()
                actor.info["PositionPatient"] = reader.GetImagePositionPatient()
                actor.info["OrientationPatient"] = reader.GetImageOrientationPatient()
                actor.info["BitsAllocated"] = reader.GetBitsAllocated()
                actor.info["PixelRepresentation"] = reader.GetPixelRepresentation()
                actor.info["NumberOfComponents"] = reader.GetNumberOfComponents()
                actor.info["TransferSyntaxUID"] = reader.GetTransferSyntaxUID()
                actor.info["RescaleSlope"] = reader.GetRescaleSlope()
                actor.info["RescaleOffset"] = reader.GetRescaleOffset()
                actor.info["PatientName"] = reader.GetPatientName()
                actor.info["StudyUID"] = reader.GetStudyUID()
                actor.info["StudyID"] = reader.GetStudyID()
                actor.info["GantryAngle"] = reader.GetGantryAngle()

                acts.append(actor)

            else:  ### it's a normal directory
                utils.humansort(flist)
                for ifile in flist:
                    a = _load_file(fod + "/" + ifile, unpack)
                    acts.append(a)
        else:
            vedo.logger.error(f"in load(), cannot find {fod}")

    if len(acts) == 1:
        if "numpy" in str(type(acts[0])):
            return acts[0]
        if not acts[0]:
            vedo.logger.error(f"in load(), cannot load {inputobj}")
        return acts[0]

    if len(acts) == 0:
        vedo.logger.error(f"in load(), cannot load {inputobj}")
        return None

    else:
        return acts


def _load_file(filename, unpack):
    fl = filename.lower()

    ################################################################# other formats:
    if fl.endswith(".xml") or fl.endswith(".xml.gz") or fl.endswith(".xdmf"):
        # Fenics tetrahedral file
        actor = loadDolfin(filename)
    elif fl.endswith(".neutral") or fl.endswith(".neu"):  # neutral tetrahedral file
        actor = loadNeutral(filename)
    elif fl.endswith(".gmsh"):  # gmesh file
        actor = loadGmesh(filename)
    elif fl.endswith(".pcd"):  # PCL point-cloud format
        actor = loadPCD(filename)
        actor.GetProperty().SetPointSize(2)
    elif fl.endswith(".off"):
        actor = loadOFF(filename)
    elif fl.endswith(".3ds"):  # 3ds format
        actor = load3DS(filename)
    elif fl.endswith(".wrl"):
        importer = vtk.vtkVRMLImporter()
        importer.SetFileName(filename)
        importer.Read()
        importer.Update()
        actors = importer.GetRenderer().GetActors()  # vtkActorCollection
        actors.InitTraversal()
        wacts = []
        for i in range(actors.GetNumberOfItems()):
            act = actors.GetNextActor()
            wacts.append(act)
        actor = Assembly(wacts)

        ################################################################# volumetric:
    elif (
        fl.endswith(".tif")
        or fl.endswith(".tiff")
        or fl.endswith(".slc")
        or fl.endswith(".vti")
        or fl.endswith(".mhd")
        or fl.endswith(".nrrd")
        or fl.endswith(".nii")
        or fl.endswith(".dem")
    ):
        img = loadImageData(filename)
        actor = Volume(img)

        ################################################################# 2D images:
    elif (
        fl.endswith(".png")
        or fl.endswith(".jpg")
        or fl.endswith(".bmp")
        or fl.endswith(".jpeg")
        or fl.endswith(".gif")
    ):
        if ".png" in fl:
            picr = vtk.vtkPNGReader()
        elif ".jpg" in fl or ".jpeg" in fl:
            picr = vtk.vtkJPEGReader()
        elif ".bmp" in fl:
            picr = vtk.vtkBMPReader()
        elif ".gif" in fl:
            from PIL import Image, ImageSequence

            img = Image.open(filename)
            frames = []
            for frame in ImageSequence.Iterator(img):
                a = np.array(frame.convert("RGB").getdata(), dtype=np.uint8)
                a = a.reshape([frame.size[1], frame.size[0], 3])
                frames.append(Picture(a))
            return frames

        picr.SetFileName(filename)
        picr.Update()
        actor = Picture(picr.GetOutput())  # object derived from vtk.vtkImageActor()

        ################################################################# multiblock:
    elif fl.endswith(".vtm") or fl.endswith(".vtmb"):
        read = vtk.vtkXMLMultiBlockDataReader()
        read.SetFileName(filename)
        read.Update()
        mb = read.GetOutput()
        if unpack:
            acts = []
            for i in range(mb.GetNumberOfBlocks()):
                b = mb.GetBlock(i)
                if isinstance(
                    b,
                    (
                        vtk.vtkPolyData,
                        vtk.vtkUnstructuredGrid,
                        vtk.vtkStructuredGrid,
                        vtk.vtkRectilinearGrid,
                    ),
                ):
                    acts.append(Mesh(b))
                elif isinstance(b, vtk.vtkImageData):
                    acts.append(Volume(b))
                elif isinstance(b, vtk.vtkUnstructuredGrid):
                    acts.append(vedo.UGrid(b))
            return acts
        return mb

        ################################################################# numpy:
    elif fl.endswith(".npy") or fl.endswith(".npz"):
        acts = loadnumpy(filename)

        if unpack is False:
            return Assembly(acts)
        return acts

    elif fl.endswith(".geojson"):
        return loadGeoJSON(filename)

    elif fl.endswith(".pvd"):
        return loadPVD(filename)

        ################################################################# polygonal mesh:
    else:
        if fl.endswith(".vtk"):  # read all legacy vtk types

            # output can be:
            # PolyData, StructuredGrid, StructuredPoints, UnstructuredGrid, RectilinearGrid
            reader = vtk.vtkDataSetReader()
            reader.ReadAllScalarsOn()
            reader.ReadAllVectorsOn()
            reader.ReadAllTensorsOn()
            reader.ReadAllFieldsOn()
            reader.ReadAllNormalsOn()
            reader.ReadAllColorScalarsOn()

        elif fl.endswith(".ply"):
            reader = vtk.vtkPLYReader()
        elif fl.endswith(".obj"):
            reader = vtk.vtkOBJReader()
        elif fl.endswith(".stl"):
            reader = vtk.vtkSTLReader()
        elif fl.endswith(".byu") or fl.endswith(".g"):
            reader = vtk.vtkBYUReader()
        elif fl.endswith(".foam"):  # OpenFoam
            reader = vtk.vtkOpenFOAMReader()
        elif fl.endswith(".pvd"):
            reader = vtk.vtkXMLGenericDataObjectReader()
        elif fl.endswith(".vtp"):
            reader = vtk.vtkXMLPolyDataReader()
        elif fl.endswith(".vts"):
            reader = vtk.vtkXMLStructuredGridReader()
        elif fl.endswith(".vtu"):
            reader = vtk.vtkXMLUnstructuredGridReader()
        elif fl.endswith(".vtr"):
            reader = vtk.vtkXMLRectilinearGridReader()
        elif fl.endswith(".pvtk"):
            reader = vtk.vtkPDataSetReader()
        elif fl.endswith(".pvtr"):
            reader = vtk.vtkXMLPRectilinearGridReader()
        elif fl.endswith("pvtu"):
            reader = vtk.vtkXMLPUnstructuredGridReader()
        elif fl.endswith(".txt") or fl.endswith(".xyz"):
            reader = vtk.vtkParticleReader()  # (format is x, y, z, scalar)
        elif fl.endswith(".facet"):
            reader = vtk.vtkFacetReader()
        else:
            return None

        reader.SetFileName(filename)
        reader.Update()
        routput = reader.GetOutput()

        if not routput:
            vedo.logger.error(f"unable to load {filename}")
            return None

        if isinstance(routput, vtk.vtkUnstructuredGrid):
            actor = vedo.TetMesh(routput)

        else:
            actor = Mesh(routput)
            if fl.endswith(".txt") or fl.endswith(".xyz"):
                actor.GetProperty().SetPointSize(4)

    actor.filename = filename
    actor.file_size, actor.created = file_info(filename)
    return actor


def download(url, force=False, verbose=True):
    """Retrieve a file from a URL, save it locally and return its path.
    Use `force` to force reload and discard cached copies."""

    if not url.startswith("https://"):
        vedo.logger.error(f"Invalid URL (must start with https):\n{url}")
        return url
    url = url.replace("www.dropbox", "dl.dropbox")

    if "github.com" in url:
        url = url.replace("/blob/", "/raw/")

    basename = os.path.basename(url)

    if "?" in basename:
        basename = basename.split("?")[0]

    tmp_file = NamedTemporaryFile(delete=False)
    tmp_file.name = os.path.join(os.path.dirname(tmp_file.name), os.path.basename(basename))

    if not force and os.path.exists(tmp_file.name):
        if verbose:
            colors.printc("reusing cached file:", tmp_file.name)
            # colors.printc("     (use force=True to force a new download)")
        return tmp_file.name

    try:
        from urllib.request import urlopen, Request

        req = Request(url, headers={"User-Agent": "Mozilla/5.0"})
        if verbose:
            colors.printc("reading", basename, "from", url.split("/")[2][:40], "...", end="")

    except ImportError:
        import urllib2
        import contextlib

        urlopen = lambda url_: contextlib.closing(urllib2.urlopen(url_))
        req = url
        if verbose:
            colors.printc("reading", basename, "from", url.split("/")[2][:40], "...", end="")

    with urlopen(req) as response, open(tmp_file.name, "wb") as output:
        output.write(response.read())

    if verbose:
        colors.printc(" done.")
    return tmp_file.name


def gunzip(filename):
    """Unzip a `.gz` file to a temporary file and returns its path."""
    if not filename.endswith(".gz"):
        # colors.printc("gunzip() error: file must end with .gz", c='r')
        return filename
    import gzip

    tmp_file = NamedTemporaryFile(delete=False)
    tmp_file.name = os.path.join(
        os.path.dirname(tmp_file.name), os.path.basename(filename).replace(".gz", "")
    )
    inF = gzip.open(filename, "rb")
    with open(tmp_file.name, "wb") as outF:
        outF.write(inF.read())
    inF.close()
    return tmp_file.name


def file_info(file_path):
    """Return the file size and creation time of input file"""
    sz, created = "", ""
    if os.path.isfile(file_path):
        file_info = os.stat(file_path)
        num = file_info.st_size
        for x in ["B", "KB", "MB", "GB", "TB"]:
            if num < 1024.0:
                break
            num /= 1024.0
        sz = "%3.1f%s" % (num, x)
        created = time.ctime(os.path.getmtime(file_path))
    return sz, created


###################################################################
def loadStructuredPoints(filename):
    """Load and return a `vtkStructuredPoints` object from file."""
    reader = vtk.vtkStructuredPointsReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadStructuredGrid(filename):
    """Load and return a `vtkStructuredGrid` object from file."""
    if filename.endswith(".vts"):
        reader = vtk.vtkXMLStructuredGridReader()
    else:
        reader = vtk.vtkStructuredGridReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadUnStructuredGrid(filename):
    """Load and return a `vtkunStructuredGrid` object from file."""
    if filename.endswith(".vtu"):
        reader = vtk.vtkXMLUnstructuredGridReader()
    else:
        reader = vtk.vtkUnstructuredGridReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadRectilinearGrid(filename):
    """Load and return a `vtkRectilinearGrid` object from file."""
    if filename.endswith(".vtr"):
        reader = vtk.vtkXMLRectilinearGridReader()
    else:
        reader = vtk.vtkRectilinearGridReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


def loadXMLData(filename):
    """Read any type of vtk data object encoded in XML format."""
    reader = vtk.vtkXMLGenericDataObjectReader()
    reader.SetFileName(filename)
    reader.Update()
    return reader.GetOutput()


###################################################################
def load3DS(filename):
    """Load `3DS` file format from file.
    Returns:
        `Assembly(vtkAssembly)` object.
    """
    renderer = vtk.vtkRenderer()
    renWin = vtk.vtkRenderWindow()
    renWin.AddRenderer(renderer)

    importer = vtk.vtk3DSImporter()
    importer.SetFileName(filename)
    importer.ComputeNormalsOn()
    importer.SetRenderWindow(renWin)
    importer.Update()

    actors = renderer.GetActors()  # vtkActorCollection
    acts = []
    for i in range(actors.GetNumberOfItems()):
        a = actors.GetItemAsObject(i)
        acts.append(a)
    del renWin
    return Assembly(acts)


def loadOFF(filename):
    """Read the OFF file format (polygonal mesh)."""
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    vertices = []
    faces = []
    NumberOfVertices = None
    i = -1
    for text in lines:
        if len(text) == 0:
            continue
        if text == "\n":
            continue
        if "#" in text:
            continue
        if "OFF" in text:
            continue

        ts = text.split()
        n = len(ts)

        if not NumberOfVertices and n > 1:
            NumberOfVertices, NumberOfFaces = int(ts[0]), int(ts[1])
            continue
        i += 1

        if i < NumberOfVertices and n == 3:
            x, y, z = float(ts[0]), float(ts[1]), float(ts[2])
            vertices.append([x, y, z])

        ids = []
        if NumberOfVertices <= i < (NumberOfVertices + NumberOfFaces + 1) and n > 2:
            ids += [int(xx) for xx in ts[1:]]
            faces.append(ids)

    return Mesh(utils.buildPolyData(vertices, faces))


def loadGeoJSON(filename):
    """Load GeoJSON files."""
    jr = vtk.vtkGeoJSONReader()
    jr.SetFileName(filename)
    jr.Update()
    return Mesh(jr.GetOutput())


def loadDolfin(filename, exterior=False):
    """Reads a `Fenics/Dolfin` file format (.xml or .xdmf).
    Return an `Mesh` object."""
    import dolfin

    if filename.lower().endswith(".xdmf"):
        f = dolfin.XDMFFile(filename)
        m = dolfin.Mesh()
        f.read(m)
    else:
        m = dolfin.Mesh(filename)

    bm = dolfin.BoundaryMesh(m, "exterior")

    if exterior:
        poly = utils.buildPolyData(bm.coordinates(), bm.cells(), tetras=True)
    else:
        polyb = utils.buildPolyData(bm.coordinates(), bm.cells(), tetras=True)
        polym = utils.buildPolyData(m.coordinates(), m.cells(), tetras=True)
        app = vtk.vtkAppendPolyData()
        app.AddInputData(polym)
        app.AddInputData(polyb)
        app.Update()
        poly = app.GetOutput()
    return Mesh(poly).lw(0.1)


def loadPVD(filename):
    """Reads a paraview set of files."""
    import xml.etree.ElementTree as et

    tree = et.parse(filename)

    dname = os.path.dirname(filename)
    if not dname:
        dname = "."

    listofobjs = []
    for coll in tree.getroot():
        for dataset in coll:
            fname = dataset.get("file")
            ob = load(dname + "/" + fname)
            tm = dataset.get("timestep")
            if tm:
                ob.time = tm
            listofobjs.append(ob)
    if len(listofobjs) == 1:
        return listofobjs[0]
    if len(listofobjs) == 0:
        return None
    return listofobjs


def loadNeutral(filename):
    """Reads a `Neutral` tetrahedral file format. Return an `Mesh` object."""
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    ncoords = int(lines[0])
    coords = []
    for i in range(1, ncoords + 1):
        x, y, z = lines[i].split()
        coords.append([float(x), float(y), float(z)])

    ntets = int(lines[ncoords + 1])
    idolf_tets = []
    for i in range(ncoords + 2, ncoords + ntets + 2):
        text = lines[i].split()
        v0, v1, v2, v3 = int(text[1]) - 1, int(text[2]) - 1, int(text[3]) - 1, int(text[4]) - 1
        idolf_tets.append([v0, v1, v2, v3])

    poly = utils.buildPolyData(coords, idolf_tets)
    return Mesh(poly)


def loadGmesh(filename):
    """Reads a `gmesh` file format. Return an `Mesh` object."""
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    nnodes = 0
    index_nodes = 0
    for i, line in enumerate(lines):
        if "$Nodes" in line:
            index_nodes = i + 1
            nnodes = int(lines[index_nodes])
            break
    node_coords = []
    for i in range(index_nodes + 1, index_nodes + 1 + nnodes):
        cn = lines[i].split()
        node_coords.append([float(cn[1]), float(cn[2]), float(cn[3])])

    nelements = 0
    index_elements = 0
    for i, line in enumerate(lines):
        if "$Elements" in line:
            index_elements = i + 1
            nelements = int(lines[index_elements])
            break
    elements = []
    for i in range(index_elements + 1, index_elements + 1 + nelements):
        ele = lines[i].split()
        elements.append([int(ele[-3]), int(ele[-2]), int(ele[-1])])

    poly = utils.buildPolyData(node_coords, elements, indexOffset=1)
    return Mesh(poly)


def loadPCD(filename):
    """Return a `Mesh` made of only vertex points
    from `Point Cloud` file format. Return an `Points` object."""
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()

    start = False
    pts = []
    N, expN = 0, 0
    for text in lines:
        if start:
            if N >= expN:
                break
            l = text.split()
            pts.append([float(l[0]), float(l[1]), float(l[2])])
            N += 1
        if not start and "POINTS" in text:
            expN = int(text.split()[1])
        if not start and "DATA ascii" in text:
            start = True
    if expN != N:
        vedo.logger.warning(f"Mismatch in PCD file {expN} != {len(pts)}")
    poly = utils.buildPolyData(pts)
    return Points(poly).pointSize(4)


def tonumpy(obj):
    """Dump a vedo object to numpy format."""

    adict = {}
    adict["type"] = "unknown"

    ########################################################
    def _fillcommon(obj, adict):
        adict["filename"] = obj.filename
        adict["name"] = obj.name
        adict["time"] = obj.time
        adict["rendered_at"] = obj.rendered_at
        adict["position"] = obj.pos()
        adict["info"] = obj.info

        try:
            # GetMatrix might not exist for non linear transforms
            m = np.eye(4)
            vm = obj.get_transform().GetMatrix()
            for i in [0, 1, 2, 3]:
                for j in [0, 1, 2, 3]:
                    m[i, j] = vm.GetElement(i, j)
            adict["transform"] = m
            minv = np.eye(4)
            vm.Invert()
            for i in [0, 1, 2, 3]:
                for j in [0, 1, 2, 3]:
                    minv[i, j] = vm.GetElement(i, j)
            adict["transform_inverse"] = minv
        except AttributeError:
            adict["transform"] = []
            adict["transform_inverse"] = []

    ########################################################
    def _fillmesh(obj, adict):

        adict["points"] = obj.points(transformed=True).astype(float)
        poly = obj.polydata()

        adict["cells"] = None
        if poly.GetNumberOfPolys():
            try:
                adict["cells"] = np.array(obj.faces(), dtype=np.uint32)
            except ValueError:  # in case of inhomogeneous shape
                adict["cells"] = obj.faces()

        adict["lines"] = None
        if poly.GetNumberOfLines():
            adict["lines"] = obj.lines()

        adict["pointdata"] = []
        for iname in obj.pointdata.keys():
            if not iname:
                continue
            if "Normals" in iname.lower():
                continue
            arr = poly.GetPointData().GetArray(iname)
            adict["pointdata"].append([utils.vtk2numpy(arr), iname])

        adict["celldata"] = []
        for iname in obj.celldata.keys():
            if not iname:
                continue
            if "Normals" in iname.lower():
                continue
            arr = poly.GetCellData().GetArray(iname)
            adict["celldata"].append([utils.vtk2numpy(arr), iname])

        adict["activedata"] = None
        if poly.GetPointData().GetScalars():
            adict["activedata"] = ["pointdata", poly.GetPointData().GetScalars().GetName()]
        elif poly.GetCellData().GetScalars():
            adict["activedata"] = ["celldata", poly.GetCellData().GetScalars().GetName()]

        adict["LUT"] = None
        adict["LUT_range"] = None
        lut = obj.mapper().GetLookupTable()
        if lut:
            nlut = lut.GetNumberOfTableValues()
            lutvals = []
            for i in range(nlut):
                v4 = lut.GetTableValue(i)  # r, g, b, alpha
                lutvals.append(v4)
            adict["LUT"] = lutvals
            adict["LUT_range"] = lut.GetRange()

        prp = obj.GetProperty()
        adict["alpha"] = prp.GetOpacity()
        adict["representation"] = prp.GetRepresentation()
        adict["pointsize"] = prp.GetPointSize()

        adict["linecolor"] = None
        adict["linewidth"] = None
        if prp.GetEdgeVisibility():
            adict["linewidth"] = obj.linewidth()
            adict["linecolor"] = prp.GetEdgeColor()

        adict["ambient"] = prp.GetAmbient()
        adict["diffuse"] = prp.GetDiffuse()
        adict["specular"] = prp.GetSpecular()
        adict["specularpower"] = prp.GetSpecularPower()
        adict["specularcolor"] = prp.GetSpecularColor()
        adict["shading"] = prp.GetInterpolation()  # flat phong..:
        adict["color"] = prp.GetColor()
        adict["lighting_is_on"] = prp.GetLighting()
        adict["backcolor"] = None
        if obj.GetBackfaceProperty():
            adict["backcolor"] = obj.GetBackfaceProperty().GetColor()

        adict["scalarvisibility"] = obj.mapper().GetScalarVisibility()
        adict["texture"] = None

    ######################################################## Assembly
    if isinstance(obj, Assembly):
        pass
        # adict['type'] = 'Assembly'
        # _fillcommon(obj, adict)
        # adict['actors'] = []
        # for a in obj.unpack():
        #     assdict = dict()
        #     if isinstance(a, Mesh):
        #         _fillmesh(a, assdict)
        #         adict['actors'].append(assdict)

    ######################################################## Points/Mesh
    elif isinstance(obj, Points):
        adict["type"] = "Mesh"
        _fillcommon(obj, adict)
        _fillmesh(obj, adict)

    ######################################################## Volume
    elif isinstance(obj, Volume):
        adict["type"] = "Volume"
        _fillcommon(obj, adict)
        imgdata = obj.inputdata()
        arr = utils.vtk2numpy(imgdata.GetPointData().GetScalars())
        adict["array"] = arr.reshape(imgdata.GetDimensions())
        adict["mode"] = obj.mode()
        # adict['jittering'] = obj.mapper().GetUseJittering()

        prp = obj.GetProperty()
        ctf = prp.GetRGBTransferFunction()
        otf = prp.GetScalarOpacity()
        gotf = prp.GetGradientOpacity()
        smin, smax = ctf.GetRange()
        xs = np.linspace(smin, smax, num=100, endpoint=True)
        cols, als, algrs = [], [], []
        for x in xs:
            cols.append(ctf.GetColor(x))
            als.append(otf.GetValue(x))
            if gotf:
                algrs.append(gotf.GetValue(x))
        adict["color"] = cols
        adict["alpha"] = als
        adict["alphagrad"] = algrs

    ######################################################## Picture
    elif isinstance(obj, Picture):
        adict["type"] = "Picture"
        _fillcommon(obj, adict)
        adict["array"] = obj.tonumpy()

    ######################################################## Text2D
    elif isinstance(obj, vedo.Text2D):
        adict["type"] = "Text2D"
        adict["rendered_at"] = obj.rendered_at
        adict["text"] = obj.text()
        adict["position"] = obj.GetPosition()
        adict["color"] = obj.property.GetColor()
        adict["font"] = obj.fontname
        adict["size"] = obj.property.GetFontSize() / 22.5
        adict["bgcol"] = obj.property.GetBackgroundColor()
        adict["alpha"] = obj.property.GetBackgroundOpacity()
        adict["frame"] = obj.property.GetFrame()
        # print('tonumpy(): vedo.Text2D', obj.text()[:10], obj.font(), obj.GetPosition())

    else:
        pass
        # colors.printc('Unknown object type in tonumpy()', [obj], c='r')

    return adict


def loadnumpy(inobj):
    """Load a vedo format file or scene."""
    # make sure the numpy file is not containing a scene
    if isinstance(inobj, str):  # user passing a file

        if inobj.endswith(".npy"):
            data = np.load(inobj, allow_pickle=True, encoding="latin1")  # .flatten()
        elif inobj.endswith(".npz"):
            data = np.load(inobj, allow_pickle=True)["vedo_scenes"]

        isdict = hasattr(data[0], "keys")

        if isdict and "objects" in data[0].keys():  # loading a full scene!!
            return import_window(data[0])

        # it's a very normal numpy data object? just return it!
        if not isdict:
            return data
        if "type" not in data[0].keys():
            return data

    else:
        data = inobj

    ######################################################
    def _load_common(obj, d):
        keys = d.keys()
        if "time" in keys:
            obj.time = d["time"]
        if "name" in keys:
            obj.name = d["name"]
        if "filename" in keys:
            obj.filename = d["filename"]
        if "info" in keys:
            obj.info = d["info"]

        # if "transform" in keys and len(d["transform"]) == 4:
        #     vm = vtk.vtkMatrix4x4()
        #     for i in [0, 1, 2, 3]:
        #         for j in [0, 1, 2, 3]:
        #             vm.SetElement(i, j, d["transform"][i, j])
        #     obj.apply_transform(vm)

        elif "position" in keys:
            obj.pos(d["position"])

    ######################################################
    def _buildmesh(d):
        keys = d.keys()

        vertices = d["points"]
        if len(vertices) == 0:
            return None

        cells = None
        if "cells" in keys:
            cells = d["cells"]

        lines = None
        if "lines" in keys:
            lines = d["lines"]

        poly = utils.buildPolyData(vertices, cells, lines)
        msh = Mesh(poly)
        _load_common(msh, d)

        prp = msh.GetProperty()
        if 'ambient' in keys:        prp.SetAmbient(d['ambient'])
        if 'diffuse' in keys:        prp.SetDiffuse(d['diffuse'])
        if 'specular' in keys:       prp.SetSpecular(d['specular'])
        if 'specularpower' in keys:  prp.SetSpecularPower(d['specularpower'])
        if 'specularcolor' in keys:  prp.SetSpecularColor(d['specularcolor'])
        if 'lighting_is_on' in keys:   prp.SetLighting(d['lighting_is_on'])
        if 'shading' in keys:        prp.SetInterpolation(d['shading'])
        if 'alpha' in keys:          prp.SetOpacity(d['alpha'])
        if 'opacity' in keys:        prp.SetOpacity(d['opacity']) # synonym
        if 'representation' in keys: prp.SetRepresentation(d['representation'])
        if 'pointsize' in keys and d['pointsize']: prp.SetPointSize(d['pointsize'])

        if 'linewidth' in keys and d['linewidth']: msh.linewidth(d['linewidth'])
        if 'linecolor' in keys and d['linecolor']: msh.linecolor(d['linecolor'])

        if 'color' in keys and d['color'] is not None:
            msh.color(d['color'])
        if 'backcolor' in keys and d['backcolor'] is not None:
            msh.backcolor(d['backcolor'])

        if "celldata" in keys:
            for csc, cscname in d["celldata"]:
                msh.celldata[cscname] = csc
        if "pointdata" in keys:
            for psc, pscname in d["pointdata"]:
                msh.pointdata[pscname] = psc

        msh.mapper().ScalarVisibilityOff()  # deactivate scalars

        if "LUT" in keys and "activedata" in keys and d["activedata"]:
            # print(d['activedata'],'', msh.filename)
            lut_list = d["LUT"]
            ncols = len(lut_list)
            lut = vtk.vtkLookupTable()
            lut.SetNumberOfTableValues(ncols)
            lut.SetRange(d["LUT_range"])
            for i in range(ncols):
                r, g, b, a = lut_list[i]
                lut.SetTableValue(i, r, g, b, a)
            lut.Build()
            msh.mapper().SetLookupTable(lut)
            msh.mapper().ScalarVisibilityOn()  # activate scalars
            msh.mapper().SetScalarRange(d["LUT_range"])
            if d["activedata"][0] == "celldata":
                poly.GetCellData().SetActiveScalars(d["activedata"][1])
            if d["activedata"][0] == "pointdata":
                poly.GetPointData().SetActiveScalars(d["activedata"][1])

        if "shading" in keys and int(d["shading"]) > 0:
            msh.compute_normals(cells=0)  # otherwise cannot renderer phong

        if "scalarvisibility" in keys:
            if d["scalarvisibility"]:
                msh.mapper().ScalarVisibilityOn()
            else:
                msh.mapper().ScalarVisibilityOff()

        if "texture" in keys and d["texture"]:
            msh.texture(d["texture"])

        return msh

    ######################################################

    objs = []
    for d in data:
        # print('loadnumpy:', d['type'], d)

        ### Mesh
        if d['type'].lower() == 'mesh':
            a = _buildmesh(d)
            if a:
                objs.append(a)

        ### Assembly
        elif d['type'].lower() == 'assembly':
            assacts = []
            for ad in d["actors"]:
                assacts.append(_buildmesh(ad))
            asse = Assembly(assacts)
            _load_common(asse, d)
            objs.append(asse)

        ### Volume
        elif d['type'].lower() == 'volume':
            vol = Volume(d["array"])
            _load_common(vol, d)
            if "jittering" in d.keys():
                vol.jittering(d["jittering"])
            # print(d['mode'])
            vol.mode(d["mode"])
            vol.color(d["color"])
            vol.alpha(d["alpha"])
            vol.alpha_gradient(d["alphagrad"])
            objs.append(vol)

        ### Picture
        elif d['type'].lower() == 'picture':
            vimg = Picture(d["array"])
            _load_common(vimg, d)
            objs.append(vimg)

        ### Text2D
        elif d['type'].lower() == 'text2d':
            t = vedo.shapes.Text2D(d["text"], font=d["font"], c=d["color"])
            t.pos(d["position"]).size(d["size"])
            t.background(d["bgcol"], d["alpha"])
            if d["frame"]:
                t.frame(d["bgcol"])
            objs.append(t)

        ### Annotation ## backward compatibility - will disappear
        elif d['type'].lower() == 'annotation':

            pos = d["position"]
            if isinstance(pos, int):
                pos = "top-left"
                d["size"] *= 2.7
            t = vedo.shapes.Text2D(d["text"], font=d["font"], c=d["color"]).pos(pos)
            t.background(d["bgcol"], d["alpha"]).size(d["size"]).frame(d["bgcol"])
            objs.append(t)  ## backward compatibility

    if len(objs) == 1:
        return objs[0]
    if len(objs) == 0:
        return None
    return objs


def loadImageData(filename):
    """Read and return a `vtkImageData` object from file."""
    if ".tif" in filename.lower():
        reader = vtk.vtkTIFFReader()
        # print("GetOrientationType ", reader.GetOrientationType())
        reader.SetOrientationType(settings.tiff_orientation_type)
    elif ".slc" in filename.lower():
        reader = vtk.vtkSLCReader()
        if not reader.CanReadFile(filename):
            vedo.logger.error(f"sorry, bad SLC file {filename}")
            return None
    elif ".vti" in filename.lower():
        reader = vtk.vtkXMLImageDataReader()
    elif ".mhd" in filename.lower():
        reader = vtk.vtkMetaImageReader()
    elif ".dem" in filename.lower():
        reader = vtk.vtkDEMReader()
    elif ".nii" in filename.lower():
        reader = vtk.vtkNIFTIImageReader()
    elif ".nrrd" in filename.lower():
        reader = vtk.vtkNrrdReader()
        if not reader.CanReadFile(filename):
            vedo.logger.error(f"sorry, bad NRRD file {filename}")
            return None
    else:
        vedo.logger.error(f"sorry, cannot read file {filename}")
        return None
    reader.SetFileName(filename)
    reader.Update()
    image = reader.GetOutput()
    return image


###########################################################
def write(objct, fileoutput, binary=True):
    """
    Write object to file.

    Possile extensions are:
        - `vtk, vti, npy, npz, ply, obj, stl, byu, vtp, vti, mhd, xyz, tif, png, bmp`
    """
    obj = objct
    if isinstance(obj, Points):  # picks transformation
        obj = objct.polydata(True)
    elif isinstance(obj, (vtk.vtkActor, vtk.vtkVolume)):
        obj = objct.GetMapper().GetInput()
    elif isinstance(obj, (vtk.vtkPolyData, vtk.vtkImageData)):
        obj = objct

    fr = fileoutput.lower()
    if fr.endswith(".vtk"):
        writer = vtk.vtkDataSetWriter()
    elif fr.endswith(".ply"):
        writer = vtk.vtkPLYWriter()
        writer.AddComment("PLY file generated by vedo")
        lut = objct.GetMapper().GetLookupTable()
        if lut:
            pscal = obj.GetPointData().GetScalars()
            if not pscal:
                pscal = obj.GetCellData().GetScalars()
            if pscal and pscal.GetName():
                writer.SetArrayName(pscal.GetName())
            writer.SetLookupTable(lut)
    elif fr.endswith(".stl"):
        writer = vtk.vtkSTLWriter()
    elif fr.endswith(".vtp"):
        writer = vtk.vtkXMLPolyDataWriter()
    elif fr.endswith(".vtu"):
        writer = vtk.vtkXMLUnstructuredGridWriter()
    elif fr.endswith(".vtm"):
        g = vtk.vtkMultiBlockDataGroupFilter()
        for ob in objct:
            if isinstance(ob, (Points, Volume)):  # picks transformation
                ob = ob.polydata(True)
                g.AddInputData(ob)
        g.Update()
        mb = g.GetOutputDataObject(0)
        wri = vtk.vtkXMLMultiBlockDataWriter()
        wri.SetInputData(mb)
        wri.SetFileName(fileoutput)
        wri.Write()
        return mb
    elif fr.endswith(".xyz"):
        writer = vtk.vtkSimplePointsWriter()
    elif fr.endswith(".facet"):
        writer = vtk.vtkFacetWriter()
    elif fr.endswith(".vti"):
        writer = vtk.vtkXMLImageDataWriter()
    elif fr.endswith(".mhd"):
        writer = vtk.vtkMetaImageWriter()
    elif fr.endswith(".nii"):
        writer = vtk.vtkNIFTIImageWriter()
    elif fr.endswith(".png"):
        writer = vtk.vtkPNGWriter()
    elif fr.endswith(".jpg"):
        writer = vtk.vtkJPEGWriter()
    elif fr.endswith(".bmp"):
        writer = vtk.vtkBMPWriter()
    elif fr.endswith(".tif") or fr.endswith(".tiff"):
        writer = vtk.vtkTIFFWriter()
        writer.SetFileDimensionality(len(obj.GetDimensions()))
    elif fr.endswith(".npy") or fr.endswith(".npz"):
        if utils.is_sequence(objct):
            objslist = objct
        else:
            objslist = [objct]
        dicts2save = []
        for obj in objslist:
            dicts2save.append(tonumpy(obj))
        np.save(fileoutput, dicts2save)
        return dicts2save

    elif fr.endswith(".obj"):
        with open(fileoutput, "w", encoding="UTF-8") as outF:
            outF.write("# OBJ file format with ext .obj\n")
            outF.write("# File generated by vedo\n")

            for p in objct.points():
                outF.write("v {:.5g} {:.5g} {:.5g}\n".format(*p))

            ptxt = objct.polydata().GetPointData().GetTCoords()
            if ptxt:
                ntxt = utils.vtk2numpy(ptxt)
                for vt in ntxt:
                    outF.write("vt " + str(vt[0]) + " " + str(vt[1]) + " 0.0\n")

            if isinstance(objct, Mesh):

                for i, f in enumerate(objct.faces()):
                    fs = ""
                    for fi in f:
                        if ptxt:
                            fs += f" {fi+1}/{fi+1}"
                        else:
                            fs += f" {fi+1}"
                    outF.write(f"f{fs}\n")

                for l in objct.lines():
                    ls = ""
                    for li in l:
                        ls += str(li + 1) + " "
                    outF.write(f"l {ls}\n")

        return objct

    elif fr.endswith(".xml"):  # write tetrahedral dolfin xml
        vertices = objct.points().astype(str)
        faces = np.array(objct.faces()).astype(str)
        ncoords = vertices.shape[0]
        with open(fileoutput, "w", encoding="UTF-8") as outF:
            outF.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            outF.write('<dolfin xmlns:dolfin="http://www.fenicsproject.org">\n')

            if len(faces[0]) == 4:  # write tetrahedral mesh
                ntets = faces.shape[0]
                outF.write('  <mesh celltype="tetrahedron" dim="3">\n')
                outF.write('    <vertices size="' + str(ncoords) + '">\n')
                for i in range(ncoords):
                    x, y, z = vertices[i]
                    outF.write('      <vertex index="'+str(i)+'" x="'+x+'" y="'+y+'" z="'+z+'"/>\n')
                outF.write('    </vertices>\n')
                outF.write('    <cells size="' + str(ntets) + '">\n')
                for i in range(ntets):
                    v0, v1, v2, v3 = faces[i]
                    outF.write('     <tetrahedron index="'+str(i)
                               + '" v0="'+v0+'" v1="'+v1+'" v2="'+v2+'" v3="'+v3+'"/>\n')

            elif len(faces[0]) == 3:  # write triangle mesh
                ntri = faces.shape[0]
                outF.write('  <mesh celltype="triangle" dim="2">\n')
                outF.write('    <vertices size="' + str(ncoords) + '">\n')
                for i in range(ncoords):
                    x, y, dummy_z = vertices[i]
                    outF.write('      <vertex index="'+str(i)+'" x="'+x+'" y="'+y+'"/>\n')
                outF.write('    </vertices>\n')
                outF.write('    <cells size="' + str(ntri) + '">\n')
                for i in range(ntri):
                    v0, v1, v2 = faces[i]
                    outF.write('     <triangle index="'+str(i)+'" v0="'+v0+'" v1="'+v1+'" v2="'+v2+'"/>\n')

            outF.write("    </cells>\n")
            outF.write("  </mesh>\n")
            outF.write("</dolfin>\n")
        return objct

    else:
        vedo.logger.error(f"Unknown format {fileoutput}, file not saved")
        return objct

    try:
        if binary:
            writer.SetFileTypeToBinary()
        else:
            writer.SetFileTypeToASCII()
    except AttributeError:
        pass

    try:
        writer.SetInputData(obj)
        writer.SetFileName(fileoutput)
        writer.Write()
    except:
        vedo.logger.error(f"could not save {fileoutput}")
    return objct


def write_transform(inobj, filename="transform.mat", comment=""):
    """
    Save a transformation for a mesh or pointcloud to ASCII file.

    Arguments:
        filename : (str)
            output file name
        comment : (str)
            some optional comment
    """
    if isinstance(inobj, Points):
        M = inobj.get_transform().GetMatrix()
    elif isinstance(inobj, vtk.vtkTransform):
        M = inobj.GetMatrix()
    elif isinstance(inobj, vtk.vtkMatrix4x4):
        M = inobj
    else:
        vedo.logger.error(f"in write_transform(), cannot understand input type {type(inobj)}")

    with open(filename, "w", encoding="UTF-8") as f:
        if comment:
            f.write("# " + comment + "\n")
        for i in range(4):
            f.write(
                str(M.GetElement(i,0))+' '+
                str(M.GetElement(i,1))+' '+
                str(M.GetElement(i,2))+' '+
                str(M.GetElement(i,3))+'\n',
            )
        f.write('\n')


def load_transform(filename):
    """
    Load a transformation from a file `.mat`.

    Returns:
        - `vtkTransform`
            The transformation to be applied to some object (`use apply_transform()`).
        - `str`, a comment string associated to this transformation file.
    """
    with open(filename, "r", encoding="UTF-8") as f:
        lines = f.readlines()
        M = vtk.vtkMatrix4x4()
        i = 0
        comment = ""
        for l in lines:
            if l.startswith("#"):
                comment = l.replace("#", "").replace("\n", "")
                continue
            vals = l.split(" ")
            if len(vals) == 4:
                for j in range(4):
                    v = vals[j].replace("\n", "")
                    M.SetElement(i, j, float(v))
                i += 1
        T = vtk.vtkTransform()
        T.SetMatrix(M)
    return (T, comment)


###############################################################################
def export_window(fileoutput, binary=False):
    """
    Exporter which writes out the rendered scene into an HTML, X3D
    or Numpy file.

    Example:
        - [export_x3d.py](https://github.com/marcomusy/vedo/tree/master/examples/other/export_x3d.py)

        Check out the HTML generated webpage [here](https://vedo.embl.es/examples/embryo.html).

        <img src='https://user-images.githubusercontent.com/32848391/57160341-c6ffbd80-6de8-11e9-95ff-7215ce642bc5.jpg' width="600"/>

    .. note::
        the rendering window can also be exported to `numpy` file `scene.npz`
        by pressing `E` keyboard at any moment during visualization.
    """
    fr = fileoutput.lower()

    ####################################################################
    if fr.endswith(".npy") or fr.endswith(".npz"):
        sdict = {}
        plt = vedo.plotter_instance
        sdict["shape"] = plt.shape
        sdict["sharecam"] = plt.sharecam
        sdict["camera"] = dict(
            pos=plt.camera.GetPosition(),
            focal_point=plt.camera.GetFocalPoint(),
            viewup=plt.camera.GetViewUp(),
            distance=plt.camera.GetDistance(),
            clipping_range=plt.camera.GetClippingRange(),
        )
        sdict["position"] = plt.pos
        sdict["size"] = plt.size
        sdict["axes"] = plt.axes
        sdict["title"] = plt.title
        sdict["backgrcol"] = colors.get_color(plt.renderer.GetBackground())
        sdict["backgrcol2"] = None
        if plt.renderer.GetGradientBackground():
            sdict["backgrcol2"] = plt.renderer.GetBackground2()
        sdict["use_depth_peeling"] = settings.use_depth_peeling
        sdict["render_lines_as_tubes"] = settings.render_lines_as_tubes
        sdict["hidden_line_removal"] = settings.hidden_line_removal
        sdict["visible_grid_edges"] = settings.visible_grid_edges
        sdict["use_parallel_projection"] = settings.use_parallel_projection
        sdict["default_font"] = settings.default_font
        sdict["objects"] = []

        allobjs = plt.get_meshes(include_non_pickables=True) + plt.get_volumes(include_non_pickables=True)
        acts2d = plt.renderer.GetActors2D()
        acts2d.InitTraversal()
        for _ in range(acts2d.GetNumberOfItems()):
            a = acts2d.GetNextItem()
            if isinstance(a, vedo.Text2D):
                allobjs.append(a)
        allobjs += plt.actors

        allobjs = list(set(allobjs))  # make sure its unique

        for a in allobjs:
            if a.GetVisibility():
                sdict["objects"].append(tonumpy(a))

        if fr.endswith(".npz"):
            np.savez_compressed(fileoutput, vedo_scenes=[sdict])
        else:
            np.save(fileoutput, [sdict])

    ####################################################################
    elif fr.endswith(".x3d"):
        obj = list(set(vedo.plotter_instance.get_meshes() + vedo.plotter_instance.actors))
        if vedo.plotter_instance.axes_instances:
            obj.append(vedo.plotter_instance.axes_instances[0])

        for a in obj:
            if isinstance(a, Mesh):
                newa = a.clone(transformed=True)
                vedo.plotter_instance.remove(a).add(newa)

            elif isinstance(a, Assembly):
                vedo.plotter_instance.remove(a)
                for b in a.unpack():
                    if b:
                        if a.name == "Axes":
                            newb = b.clone(transformed=True)
                        else:
                            # newb = b.clone(transformed=True) # BUG??

                            newb = b.clone(transformed=False)
                            tt = vtk.vtkTransform()
                            tt.Concatenate(a.GetMatrix())
                            tt.Concatenate(b.GetMatrix())
                            newb.PokeMatrix(vtk.vtkMatrix4x4())
                            newb.SetUserTransform(tt)

                        vedo.plotter_instance.add(newb)

        vedo.plotter_instance.render()

        exporter = vtk.vtkX3DExporter()
        exporter.SetBinary(binary)
        exporter.FastestOff()
        exporter.SetInput(vedo.plotter_instance.window)
        exporter.SetFileName(fileoutput)
        # exporter.WriteToOutputStringOn() # see below
        exporter.Update()
        exporter.Write()

        # this can reduce the size by more than half...
        #        outstring = exporter.GetOutputString().decode("utf-8") # this fails though
        #        from vedo.utils import isInteger, isNumber, precision
        #        newlines = []
        #        for l in outstring.splitlines(True):
        #            ls = l.lstrip()
        #            content = ls.split()
        #            newls = ""
        #            for c in content:
        #                c2 = c.replace(',','')
        #                if isNumber(c2) and not isInteger(c2):
        #                    newc = precision(float(c2), 4)
        #                    if ',' in c:
        #                        newls += newc + ','
        #                    else:
        #                        newls += newc + ' '
        #                else:
        #                    newls += c + ' '
        #        newlines.append(newls.lstrip()+'\n')
        #        with open("fileoutput", 'w', encoding='UTF-8') as f:
        #            l = "".join(newlines)
        #            f.write(l)

        x3d_html = _x3d_html.replace("~fileoutput", fileoutput)
        wsize = vedo.plotter_instance.window.GetSize()
        x3d_html = x3d_html.replace("~width", str(wsize[0]))
        x3d_html = x3d_html.replace("~height", str(wsize[1]))
        with open(fileoutput.replace(".x3d", ".html"), "w", encoding="UTF-8") as outF:
            outF.write(x3d_html)
            vedo.logger.info(f"Saved files {fileoutput} and {fileoutput.replace('.x3d','.html')}")

    ####################################################################
    elif fr.endswith(".html"):
        savebk = vedo.notebook_backend
        vedo.notebook_backend = "k3d"
        vedo.settings.default_backend = "k3d"
        plt = vedo.backends.get_notebook_backend(vedo.plotter_instance.actors)

        with open(fileoutput, "w", encoding="UTF-8") as fp:
            fp.write(plt.get_snapshot())

        vedo.notebook_backend = savebk
        vedo.settings.default_backend = savebk

    else:
        vedo.logger.error(f"export extension {fr.split('.')[-1]} is not supported")

    return vedo.plotter_instance


def import_window(fileinput, mtl_file=None, texture_path=None):
    """Import a whole scene from a Numpy or OBJ wavefront file.

    Arguments:
        mtl_file : (str)
            MTL file for OBJ wavefront files
        texture_path : (str)
            path of the texture files directory

    Returns:
        `Plotter` instance
    """
    data = None
    if isinstance(fileinput, dict):
        data = fileinput
    elif fileinput.endswith(".npy"):
        data = np.load(fileinput, allow_pickle=True, encoding="latin1").flatten()[0]
    elif fileinput.endswith(".npz"):
        data = np.load(fileinput, allow_pickle=True)["vedo_scenes"][0]

    if data is not None:
        if "render_lines_as_tubes" in data.keys():
            settings.render_lines_as_tubes = data["render_lines_as_tubes"]
        if "hidden_line_removal" in data.keys():
            settings.hidden_line_removal = data["hidden_line_removal"]
        if "visible_grid_edges" in data.keys():
            settings.visible_grid_edges = data["visible_grid_edges"]
        if "use_parallel_projection" in data.keys():
            settings.use_parallel_projection = data["use_parallel_projection"]
        if "use_polygon_offset" in data.keys():
            settings.use_polygon_offset = data["use_polygon_offset"]
        if "polygon_offset_factor" in data.keys():
            settings.polygon_offset_factor = data["polygon_offset_factor"]
        if "polygon_offset_units" in data.keys():
            settings.polygon_offset_units = data["polygon_offset_units"]
        if "interpolate_scalars_before_mapping" in data.keys():
            settings.interpolate_scalars_before_mapping = data["interpolate_scalars_before_mapping"]
        if "default_font" in data.keys():
            settings.default_font = data["default_font"]
        if "use_depth_peeling" in data.keys():
            settings.use_depth_peeling = data["use_depth_peeling"]

        axes = data.pop("axes", 4)
        title = data.pop("title", "")
        backgrcol = data.pop("backgrcol", "white")
        backgrcol2 = data.pop("backgrcol2", None)
        cam = data.pop("camera", None)

        if data["shape"] != (1, 1):
            data["size"] = "auto"  # disable size

        plt = vedo.Plotter(
            size=data["size"],  # not necessarily a good idea to set it
            # shape=data['shape'], # will need to create a Renderer class first
            axes=axes,
            title=title,
            bg=backgrcol,
            bg2=backgrcol2,
        )

        if cam:
            if "pos" in cam.keys():
                plt.camera.SetPosition(cam["pos"])
            if "focalPoint" in cam.keys():
                plt.camera.SetFocalPoint(cam["focalPoint"])
            if "focal_point" in cam.keys():
                plt.camera.SetFocalPoint(cam["focal_point"])
            if "viewup" in cam.keys():
                plt.camera.SetViewUp(cam["viewup"])
            if "distance" in cam.keys():
                plt.camera.SetDistance(cam["distance"])
            if "clippingRange" in cam.keys():
                plt.camera.SetClippingRange(cam["clippingRange"])
            if "clipping_range" in cam.keys():
                plt.camera.SetClippingRange(cam["clipping_range"])
            plt.resetcam = False

        if "objects" in data.keys():
            objs = loadnumpy(data["objects"])
            if not utils.is_sequence(objs):
                objs = [objs]
        else:
            # colors.printc("Trying to import a single mesh.. use load() instead.", c='r')
            # colors.printc(" -> try to load a single object with load().", c='r')
            objs = [loadnumpy(fileinput)]

        plt.actors = objs
        plt.add(objs)
        return plt

    elif ".obj" in fileinput.lower():

        window = vtk.vtkRenderWindow()
        window.SetOffScreenRendering(1)
        renderer = vtk.vtkRenderer()
        window.AddRenderer(renderer)

        importer = vtk.vtkOBJImporter()
        importer.SetFileName(fileinput)
        if mtl_file is not False:
            if mtl_file is None:
                mtl_file = fileinput.replace(".obj", ".mtl").replace(".OBJ", ".MTL")
            importer.SetFileNameMTL(mtl_file)
        if texture_path is not False:
            if texture_path is None:
                texture_path = fileinput.replace(".obj", ".txt").replace(".OBJ", ".TXT")
            importer.SetTexturePath(texture_path)
        importer.SetRenderWindow(window)
        importer.Update()

        plt = vedo.Plotter()

        actors = renderer.GetActors()
        actors.InitTraversal()
        for _ in range(actors.GetNumberOfItems()):
            vactor = actors.GetNextActor()
            act = Mesh(vactor)
            act_tu = vactor.GetTexture()
            if act_tu:
                act.texture(act_tu)
            plt.actors.append(act)
        return plt

    return None


##########################################################
def screenshot(filename="screenshot.png", scale=1, asarray=False):
    """
    Save a screenshot of the current rendering window.

    Arguments:
        scale : (int)
            Set image magnification as an integer multiplicative factor.
            E.g. setting a magnification of 2 produces an image twice as large,
            but 10x slower to generate.
        asarray : (bool)
            Return a numpy array of the image
    """
    if not vedo.plotter_instance or not vedo.plotter_instance.window:
        # vedo.logger.error("in screenshot(), rendering window is not present, skip.")
        return vedo.plotter_instance  ##########

    if asarray and scale == 1:
        nx, ny = vedo.plotter_instance.window.GetSize()
        arr = vtk.vtkUnsignedCharArray()
        vedo.plotter_instance.window.GetRGBACharPixelData(0, 0, nx-1, ny-1, 0, arr)
        narr = vedo.vtk2numpy(arr).T[:3].T.reshape([ny, nx, 3])
        narr = np.flip(narr, axis=0)
        return narr  ##########

    filename = str(filename)

    if filename.endswith(".pdf"):
        writer = vtk.vtkGL2PSExporter()
        writer.SetRenderWindow(vedo.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToPDF()
        writer.SetFilePrefix(filename.replace(".pdf", ""))
        writer.Write()
        return vedo.plotter_instance  ##########

    elif filename.endswith(".svg"):
        writer = vtk.vtkGL2PSExporter()
        writer.SetRenderWindow(vedo.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToSVG()
        writer.SetFilePrefix(filename.replace(".svg", ""))
        writer.Write()
        return vedo.plotter_instance  ##########

    elif filename.endswith(".eps"):
        writer = vtk.vtkGL2PSExporter()
        writer.SetRenderWindow(vedo.plotter_instance.window)
        writer.Write3DPropsAsRasterImageOff()
        writer.SilentOn()
        writer.SetSortToBSP()
        writer.SetFileFormatToEPS()
        writer.SetFilePrefix(filename.replace(".eps", ""))
        writer.Write()
        return vedo.plotter_instance  ##########

    if settings.screeshot_large_image:
        w2if = vtk.vtkRenderLargeImage()
        w2if.SetInput(vedo.plotter_instance.renderer)
        w2if.SetMagnification(scale)
    else:
        w2if = vtk.vtkWindowToImageFilter()
        w2if.SetInput(vedo.plotter_instance.window)
        if hasattr(w2if, "SetScale"):
            w2if.SetScale(int(scale), int(scale))
        if settings.screenshot_transparent_background:
            w2if.SetInputBufferTypeToRGBA()
        w2if.ReadFrontBufferOff()  # read from the back buffer
    w2if.Update()

    if asarray and scale != 1:
        pd = w2if.GetOutput().GetPointData()
        npdata = utils.vtk2numpy(pd.GetArray("ImageScalars"))
        npdata = npdata[:, [0, 1, 2]]
        ydim, xdim, _ = w2if.GetOutput().GetDimensions()
        npdata = npdata.reshape([xdim, ydim, -1])
        npdata = np.flip(npdata, axis=0)
        return npdata

    if filename.lower().endswith(".png"):
        writer = vtk.vtkPNGWriter()
        writer.SetFileName(filename)
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    elif filename.lower().endswith(".jpg") or filename.lower().endswith(".jpeg"):
        writer = vtk.vtkJPEGWriter()
        writer.SetFileName(filename)
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    else:  # add .png
        writer = vtk.vtkPNGWriter()
        writer.SetFileName(filename + ".png")
        writer.SetInputData(w2if.GetOutput())
        writer.Write()
    return vedo.plotter_instance


def ask(*question, **kwarg):
    """
    Ask a question from command line. Return the answer as a string.
    See function `colors.printc()` for the description of the keyword options.

    Arguments:
        options : (list)
            a python list of possible answers to choose from.
        default : (str)
            the default answer when just hitting return.

    Example:
        ```python
        import vedo
        res = vedo.file_io.ask("Continue?", options=['Y','n'], default='Y', c='g')
        print(res)
        ```
    """
    kwarg.update({"end": " "})
    if "invert" not in kwarg:
        kwarg.update({"invert": True})
    if "box" in kwarg:
        kwarg.update({"box": ""})

    options = kwarg.pop("options", [])
    default = kwarg.pop("default", "")
    if options:
        opt = "["
        for o in options:
            opt += o + "/"
        opt = opt[:-1] + "]"
        colors.printc(*question, opt, **kwarg)
    else:
        colors.printc(*question, **kwarg)

    resp = input()

    if options:
        if resp not in options:
            if default and str(repr(resp)) == "''":
                return default
            colors.printc("Please choose one option in:", opt, italic=True, bold=False)
            kwarg["options"] = options
            return ask(*question, **kwarg)  # ask again
    return resp


##############################################################################################
class Video:
    """
    Generate a video from a rendering window.
    """

    def __init__(self, name="movie.mp4", duration=None, fps=24, backend="imageio"):
        """
        Class to generate a video from the specified rendering window.
        Program `ffmpeg` is used to create video from each generated frame.

        Arguments:
            name : (str)
                name of the output file.
            fps : (int)
                set the number of frames per second.
            duration : (float)
                set the total `duration` of the video and recalculates `fps` accordingly.
            backend : (str)
                the backend engine to be used `['imageio', 'ffmpeg', 'cv']`

        Examples:
            - [make_video.py](https://github.com/marcomusy/vedo/tree/master/examples/other/make_video.py)

            ![](https://user-images.githubusercontent.com/32848391/50739007-2bfc2b80-11da-11e9-97e6-620a3541a6fa.jpg)
        """
        self.name = name
        self.duration = duration
        self.backend = backend
        self.fps = float(fps)
        self.command = "ffmpeg -loglevel panic -y -r"
        self.options = "-b:v 8000k"

        self.frames = []
        self.tmp_dir = TemporaryDirectory()
        self.get_filename = lambda x: os.path.join(self.tmp_dir.name, x)
        colors.printc(":video: Video file", self.name, "is open... ", c="m", end="")

    def add_frame(self):
        """Add frame to current video."""
        fr = self.get_filename(str(len(self.frames)) + ".png")
        screenshot(fr)
        self.frames.append(fr)
        return self

    def pause(self, pause=0):
        """Insert a `pause`, in seconds."""
        fr = self.frames[-1]
        n = int(self.fps * pause)
        for _ in range(n):
            fr2 = self.get_filename(str(len(self.frames)) + ".png")
            self.frames.append(fr2)
            os.system("cp -f %s %s" % (fr, fr2))
        return self

    def action(self, elevation=(0, 80), azimuth=(0, 359), cameras=(), resetcam=False):
        """
        Automatic shooting of a static scene by specifying rotation and elevation ranges.

        Arguments:
            elevation : list
                initial and final elevation angles
            azimuth_range : list
                initial and final azimuth angles
            cameras : list
                list of cameras to go through, each camera can be dictionary or a vtkCamera
        """
        if not self.duration:
            self.duration = 5

        plt = vedo.plotter_instance
        n = int(self.fps * self.duration)

        cams = []
        for cm in cameras:
            cams.append(utils.camera_from_dict(cm))
        nc = len(cams)

        plt.show(resetcam=resetcam, interactive=False)

        if nc:
            for i in range(n):
                plt.move_camera(cams, i / n)
                plt.show()
                self.add_frame()

        else:  ########################################

            for i in range(n):
                plt.camera.Elevation((elevation[1] - elevation[0]) / n)
                plt.camera.Azimuth((azimuth[1] - azimuth[0]) / n)
                plt.show()
                self.add_frame()

        return self

    def close(self):
        """
        Render the video and write it to file.
        """
        if self.duration:
            self.fps = int(len(self.frames) / float(self.duration) + 0.5)
            colors.printc("recalculated fps:", self.fps, c="m", end="")
        else:
            self.fps = int(self.fps)

        ########################################
        if self.backend == "ffmpeg":
            out = os.system(
                self.command
                + " "
                + str(self.fps)
                + " -i "
                + f"'{self.tmp_dir.name}'"
                + os.sep
                + "%01d.png "
                + self.options
                + " "
                + f"'{self.name}'"
            )
            if out:
                vedo.logger.error(f":noentry: backend {self.backend} returning error: {out}")
            else:
                colors.printc(f":save: saved to {self.name}", c="m")

        ########################################
        elif "cv" in self.backend:
            try:
                import cv2
            except ImportError:
                vedo.logger.error("opencv is not installed")
                return

            cap = cv2.VideoCapture(os.path.join(self.tmp_dir.name, "%1d.png"))
            fourcc = cv2.VideoWriter_fourcc(*"mp4v")
            w, h = vedo.plotter_instance.window.GetSize()
            writer = cv2.VideoWriter(self.name, fourcc, self.fps, (w, h), True)

            while True:
                ret, frame = cap.read()
                if not ret:
                    break
                writer.write(frame)

            cap.release()
            writer.release()

        ########################################
        elif "imageio" in self.backend:
            try:
                import imageio
            except ImportError:
                vedo.logger.error("Please install imageio with:\n pip install imageio[ffmpeg]")
                return

            if self.name.endswith(".mp4"):
                writer = imageio.get_writer(self.name, fps=self.fps)
            elif self.name.endswith(".gif"):
                writer = imageio.get_writer(self.name, mode="I", duration=1 / self.fps)
            elif self.name.endswith(".webm"):
                writer = imageio.get_writer(self.name, format="webm", fps=self.fps)
            else:
                vedo.logger.error(f"Unknown format of {self.name}.")
                return

            for f in utils.humansort(self.frames):
                image = imageio.v3.imread(f)
                writer.append_data(image)
            try:
                writer.close()
                colors.printc(f"... saved as {self.name}", c="m")
            except:
                colors.printc(f":noentry: Could not save video {self.name}", c="r")

        # finalize cleanup
        self.tmp_dir.cleanup()

    def split_frames(self, output_dir="video_frames", prefix="frame_", format="png"):
        """Split an existing video file into frames."""
        try:
            import imageio
        except ImportError:
            vedo.logger.error("\nPlease install imageio with:\n pip install imageio")
            return

        # Create the output directory if it doesn't exist
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        # Create a reader object to read the video
        reader = imageio.get_reader(self.name)

        # Loop through each frame of the video and save it as image
        print()
        for i, frame in utils.progressbar(
            enumerate(reader), title=f"writing {format} frames", c="m", width=20
        ):
            output_file = os.path.join(output_dir, f"{prefix}{str(i).zfill(5)}.{format}")
            imageio.imwrite(output_file, frame, format=format)



"""Drag the red points to modify the wire path
Press "b" to compute the magnetic field"""
import numpy as np
from vedo import settings, mag, utils
from vedo import Arrows, Points, StreamLines, Axes, Plotter, Text2D, Circle

def func(evt):
    if evt.keypress != "b":
        return
    txt.text("..computing field in space, please wait!")
    txt.c('red5').background('yellow7')
    plt.render()

    pts = sptool.spline().points() # extract the current spline
    field = []
    for probe in probes:
        B = np.zeros(3)
        for p0,p1 in zip(pts, np.roll(pts,1, axis=0)):
            p = (p0+p1)/2
            r = mag(p-probe)
            B += np.cross(p1-p0, p-probe)/r**3  # Biot-Savart law
        B /= max(1, mag(B))  # clamp the field magnitude near the wire
        field.append(B)
    field = np.array(field)

    arrows = Arrows(probes, probes+field/5).c('black')
    txt.text(__doc__).c('black').background(None)

    ppts1 = Points(probes)
    ppts1.pointdata["BField"] = field
    domain = ppts1.tovolume(n=4, dims=(10,10,10)) # interpolate

    ppts2 = ppts1.clone()  # make a copy
    ppts2.pointdata["BFieldIntensity"] = mag(field*255/3).astype(np.uint8)
    vol = ppts2.tovolume(n=4, dims=(10,10,10)).crop(back=0.5)
    isos = vol.isosurface(np.arange(0,250, 12)).smooth()
    isos.cmap("rainbow").lighting('off').alpha(0.5).add_scalarbar()
    isos.name = "Isosurfaces"

    streamlines = StreamLines(
    	domain,
    	probes,
        max_propagation=0.5,
        initial_step_size=0.01,
        direction="both",
    )
    streamlines.c('black').linewidth(2)
    plt.remove("Arrows", "StreamLines", "Isosurfaces", "Axes")
    plt.add(arrows, streamlines, isos,  Axes(streamlines))


probes = utils.pack_spheres([-2,2, -2,2, -2,2], radius=0.7)

settings.use_depth_peeling = True
settings.multi_samples = 0

plt = Plotter()
plt.add_callback("key press", func)

txt = Text2D(__doc__, font="Kanopus")
plt += txt

# Create a set of points in space to form a spline
circle = Circle(res=8)  # resolution = 8 points
sptool = plt.add_spline_tool(circle, pc='red', lw=4, closed=True)

plt.show().close()



"""The Lorenz attractor is a set of chaotic solutions of
a particular system of ordinary differential equations"""
from vedo import *

p = (25, -10, -7)  # starting point (initial condition)
dt = 0.01          # time step

# Define the ODE system to integrate (Lorenz equations)
pts, vel = [], []
for t in np.arange(0, 20, dt):
    x, y, z = p
    v = np.array([-8/3*x+y*z, -10*(y-z), -y*x+28*y-z])
    p = p + v * dt
    pts.append(p)
    vel.append(mag(v))

# Plot the trajectory in 3D space
line = Line(pts).lw(2).cmap("winter", vel).add_scalarbar("speed")
line.add_shadow("x",   3, alpha=0.2)
line.add_shadow("z", -25, alpha=0.2)
pt = Point(pts[0], c="red4", r=12).add_trail(lw=4)

def loop_func(event):
    global i
    if i < len(pts):
        pt.pos(pts[i]).update_trail() # move the point
        plt.render()
        i += 1
    else:
        plt.timer_callback("stop", tid) # stop the timer

i = 0
plt = Plotter(axes=dict(xygrid=False))
plt.show(line, pt, __doc__, viewup="z", interactive=False)
plt.add_callback("timer", loop_func)
tid = plt.timer_callback("start")
plt.interactive().close()




from vedo import dataurl, Picture 
from vedo.applications import SplinePlotter  # ready to use class!

pic = Picture(dataurl + "images/embryo.jpg")

plt = SplinePlotter(pic)
plt.show(mode="image", zoom='tightest')
print("Npts =", len(plt.cpoints), "NSpline =", plt.line.npoints)


#####################################################################
# This is a simplified version of vedo.applications.SplinePlotter 
#####################################################################
# from vedo import printc, precision, Plotter, Spline, Points, Text2D
#
# class MySplinePlotter(Plotter):
#     def __init__(self, **kwargs):
#         super().__init__(**kwargs)
#         self.cpoints = []
#         self.points = None
#         self.spline = None

#     def on_left_click(self, evt):
#         if not evt.actor:
#             return
#         p = evt.picked3d + [0, 0, 1]
#         self.cpoints.append(p)
#         self.update()
#         printc("Added point:", precision(p[:2], 4), c="g")

#     def on_right_click(self, evt):
#         if evt.actor and len(self.cpoints) > 0:
#             self.cpoints.pop()  # pop removes the last point
#             self.update()
#             printc("Deleted last point", c="r")

#     def on_key_press(self, evt):
#         if evt.keypress == "c":
#             self.cpoints = []
#             self.remove(self.spline, self.points).render()
#             printc("==== Cleared all points ====", c="r", invert=True)

#     def update(self):
#         self.remove([self.spline, self.points])  # remove old points and spline
#         self.points = Points(self.cpoints).ps(10).c("purple5")
#         self.points.pickable(False)  # avoid picking the same point
#         if len(self.cpoints) > 2:
#             self.spline = Spline(self.cpoints, closed=False).c("yellow5").lw(3)
#             self.add(self.points, self.spline)
#         else:
#             self.add(self.points)

# plt = MySplinePlotter(axes=True, bg="blackboard")
# plt.add_callback("key press", plt.on_key_press)
# plt.add_callback("left mouse click", plt.on_left_click)
# plt.add_callback("right mouse click", plt.on_right_click)
# plt.show(pic, mode="image", zoom=1.2)
# plt.close()



"""Histogram (or plot) in 3D.
The size of each cube is proportional to the value at that point"""
import numpy as np
from vedo import Volume, Cube, Glyph, show

# Make up some arbitrary data
X, Y, Z = np.mgrid[:10, :6, :4]
counts = 50 - ( (X-4)**2 + (Y-4)**2 + (Z-4)**2 )

# This is now a point cloud with an associated array of counts
pcloud = Volume(counts).topoints()

marker = Cube().scale(0.015)
glyphed_pcl = Glyph(pcloud, marker, scale_by_scalar=True)
glyphed_pcl.cmap('seismic').add_scalarbar('counts')

show(glyphed_pcl, __doc__, axes=1)



"""Plotting functions with error bands"""
from vedo import np, Rectangle, Text3D, Marker, Line
from vedo.pyplot import plot

# Make up same dummy data
x = np.arange(0, 6, 0.05)
y = 2+2*np.sin(2*x)/(x+1)
ye= y**2 / 10
miny = np.min(y-ye)
idx = np.argmax(y)

# Plot the two variables, return a Plot(Assembly) object:
fig = plot(
    x, y,
    yerrors=ye,
    xtitle='time in :museconds',
    ytitle='y oscillation [a.u.]',
    ylim=(0.5, 5),
    aspect=5/3,      # plot aspect ratio (xsize/ysize)
    error_band=True, # join errors on y into an error band
    lc="red2",       # line color
    ec="red7",       # error band color
    padding=0,       # no extra spaces around the content
    grid=0,          # no background grid
    axes=dict(axes_linewidth=2, xyframe_line=0),
)

# Add a grey transparent rectangle to represent an exclusion region:
fig += Rectangle([1,0.5], [2.7,5], c='grey5').lighting('off')

# Add some text (set z=2 so it stays on top):
fig += Text3D("Excluded\ntime range!",
              s=0.2, c='k', font="Quikhand").rotate_z(20).pos(1.3,3.6)

# Add a star marker at maximum of function (set z=0.1, so it stays on top):
fig += Marker('*', c='blue4').pos(x[idx], y[idx], 0.1)

# Add a dashed line to indicate the minimum
fig += Line((x[0], miny), (x[-1], miny)).pattern('- . -').lw(3)

fig.show(zoom='tight', mode='image', size=(900,600)).close()



"""Make a static 2D copy of a mesh
and place it in the rendering window"""
from vedo import Mesh, dataurl, show

man3d = Mesh(dataurl+'man.vtk').rotate_z(20).rotate_x(-70).scale(0.2)
man3d.c('darkgreen').lighting('glossy')

# Make a 2D snapshot of a 3D mesh
# The coordinate system options are
#     0. Displays
#     1. Normalized Display
#     2. Viewport (origin is the bottom-left corner of the window)
#     3. Normalized Viewport
#     4. View (origin is the center of the window)
#     5. World (anchor the 2d image to mesh)
# (returns a vtkActor2D)

man2d = man3d.clone2d(pos=[0.4,0.4], coordsys=4, c='r', alpha=1)

show(man3d, man2d, __doc__, axes=1).close()




"""Mesh a line contour with quads of variable resolution"""
from vedo import Spline, Grid, show
import numpy as np

pts = [
    [0.0, 0.0],
    [1.0, 0.0],
    [1.1, 4.0],
    [1.0, 1.5],
    [0.2, 5.0],
    [-1., 3.0],
    [0.4, 2.7],
    [-1., 2.4],
]

shape = Spline(pts, closed=True).color('red4').linewidth(5)

xcoords = np.arange(-2.0, 2.5, 0.075)
ycoords = np.arange(-0.5, 5.5, 0.075)

xcoords += np.cos(xcoords+0.6)*0.75 # make quads shrink and stretch
ycoords += np.sin(ycoords+0.5)*0.75 # to refine mesh resolution

grd = Grid(s=[xcoords, ycoords])    # create a gridded plane

msh = shape.generate_mesh(grid=grd, quads=True)

show(shape, msh, __doc__, axes=1).close()




"""Draw the shadows and trailing lines of 2 planes. Not really
a simulation.. just a way to illustrate how to move objects around!"""
from vedo import *

world = Box([0,0,0], 30, 16, 8).wireframe()

plane1 = Mesh(dataurl+"cessna.vtk").c("green")
plane1.pos(-15, 2, 0.15).add_trail(n=100)
plane1.add_shadow('z', -4).add_shadow('y', 8)

plane2 = plane1.clone().c("tomato")
plane2.pos(-15,-2,-0.20).add_trail(n=100)
plane2.add_shadow('z', -4).add_shadow('y', 8)

# Setup the scene
plt = Plotter(interactive=False)
plt.show(world, plane1, plane2, __doc__, viewup="z")

for t in np.arange(0, 3.2, 0.04):
    plane1.pos(9*t-15, 2-t, sin(3-t)).rotate_x(0+t) # make up some movement
    plane2.pos(8*t-15, t-2, sin(t-3)).rotate_x(2-t) # for the 2 planes
    plane1.update_trail().update_shadows()
    plane2.update_trail().update_shadows()
    plt.render()

plt.interactive().close()



"""Simulation of a gyroscope hanging from a spring"""
# (adapted by M. Musy from Bruce Sherwood, 2009)
from vedo import *

# ############################################################ parameters
dt = 0.005  # time step
ks = 15     # spring stiffness
Lrest = 1   # unstretched length of spring
Ls = 1      # length of gyroscope shaft
M = 1       # mass of gyroscope (massless shaft)
R = 0.4     # radius of gyroscope rotor
omega = 50  # angular velocity of rotor (rad/s, not shown)
gpos = vector(0, 0, 0)  # initial position of spring free end

# ############################################################ inits
top = vector(0, 2, 0)       # where top of spring is held
precess = vector(0, 0, 0)   # initial momentum of center of mass
Fgrav = vector(0, -M * 9.81, 0)
gaxis = vector(0, 0, 1)     # initial orientation of gyroscope
gaxis = versor(gaxis)
I = 1 / 2 * M * R ** 2      # moment of inertia of gyroscope
Lrot = I * omega * gaxis    # angular momentum
cm = gpos + 0.5 * Ls * gaxis  # center of mass of shaft

# ############################################################ the scene
plt = Plotter()
plt += __doc__

shaft = Cylinder([[0, 0, 0], Ls * gaxis], r=0.03, c="dg")
rotor = Cylinder([(Ls - 0.55) * gaxis, (Ls - 0.45) * gaxis], r=R, c="t")
bar   = Cylinder([Ls*gaxis/2-R*vector(0,1,0), Ls*gaxis/2+R*vector(0,1,0)], r=R/6, c="r")
gyro = shaft + rotor + bar  # group meshes into a single one of type Assembly

spring = Spring(top, gpos, r1=0.06, thickness=0.01, c="gray")
plt += [gyro, spring]       # add it to Plotter.
plt += Box(top, length=0.2, width=0.02, height=0.2, c="gray")
plt += Box(pos=(0, 0.5, 0), length=2.6, width=3, height=2.6, c="gray", alpha=0.2).wireframe()

# ############################################################ the physics
def loop_func(event):
    global  t, Lrot, precess, cm, gpos
    t += dt
    Fspring = -ks * versor(gpos - top) * (mag(gpos - top) - Lrest)
    torque = cross(-1/2 * Ls * versor(Lrot), Fspring)  # torque about center of mass
    Lrot += torque * dt
    precess += (Fgrav + Fspring) * dt  # momentum of center of mass
    cm += (precess / M) * dt
    gpos = cm - 1/2 * Ls * versor(Lrot)

    # set orientation along gaxis and rotate it around its axis by omega*t degrees
    gyro.orientation(Lrot, rotation=omega*t, rad=True).pos(gpos)
    spring.stretch(top, gpos)
    plt.render()

t = 0
plt.add_callback("timer", loop_func)
plt.timer_callback("start")
plt.show().close()



"""Example plot of 2 images containing an
alpha channel for modulating the opacity"""
#Credits: https://github.com/ilorevilo
from vedo import Picture, show
import numpy as np

rgbaimage1 = np.random.rand(50, 50, 4) * 255
alpharamp = np.linspace(0, 255, 50).astype(int)
rgbaimage1[:, :, 3] = alpharamp
rgbaimage2 = np.random.rand(50, 50, 4) * 255
rgbaimage2[:, :, 3] = alpharamp[::-1]

p1 = Picture(rgbaimage1, channels=4)

p2 = Picture(rgbaimage2, channels=4).z(12)

show(p1, p2, __doc__, axes=7, viewup="z").close()

# Second example: a b&w image from a numpy array
img = np.zeros([512,512])
img[0:256, 0:256] =   0
img[0:256,  256:] =  64
img[256:,  0:256] = 128
img[256:,   256:] = 255
img = img.transpose(1,0)

pict = Picture(img)

show(pict, mode="image", bg=(0.4,0.5,0.6), axes=1).close()



"""Streamlines originating from a set of seed points in space
subjected to a vectorial field defined on a small set of points.
This field is interpolated to a user-defined bounding box."""
from vedo import *
import pandas as pd

data = "https://raw.githubusercontent.com/plotly/datasets/master/vortex.csv"
df = pd.read_csv(data)
pts  = np.c_[df['x'], df['y'], df['z']]
wind = np.c_[df['u'], df['v'], df['w']]

domain = Points(pts)
domain.pointdata["Wind"] = wind

seeds = domain.clone().subsample(0.2) # these are the seed points

# Compute stream lines with Runge-Kutta integration, we
# extrapolate the field defined on points to a bounding box
streamlines = StreamLines(
    domain,
    seeds,
    max_propagation=100,
    extrapolate_to_box=dict(bounds=[-20,20, -15,15, -20,20]),
)
streamlines.lw(5).cmap("Blues", "Wind").add_scalarbar()

show(streamlines, __doc__, axes=1, viewup='z').close()



"""Draw the ellipse and the ellipsoid that contain 50% of a pointcloud,
then check how many points are inside both objects"""
from vedo import *

pts = Points(np.random.randn(1000,3))
pts.rotate_z(30).scale([1.5, 2, 0.01]).pos(2,3,0)

elli2d = pca_ellipse(  pts, pvalue=0.5)
elli3d = pca_ellipsoid(pts, pvalue=0.5)

extruded = elli2d.z(-0.1).extrude(0.2)  # make an oval box

printc("Inside ellipse  :", extruded.inside_points(pts).npoints, c='b')
printc("Inside ellipsoid:", elli3d.inside_points(pts).npoints, c='b')

show(pts, elli2d, elli3d, __doc__, axes=1)




"""Smoothing a mesh"""
from vedo import dataurl, Mesh, show

s1 = Mesh(dataurl+'panther.stl').lw(0.1)

s2 = s1.clone().x(50)  # place at x=50
s2.subdivide(3).smooth().compute_normals()
s2.c('light blue').lw(0).lighting('glossy').phong()

# other useful filters to combine are
# mesh.decimate(), clean(), smooth()

cam = dict(
    position=(113, -189, 62.1),
    focal_point=(18.3, 4.39, 2.41),
    viewup=(-0.0708, 0.263, 0.962),
    distance=223,
)
show(s1, s2, __doc__,
     bg='black', bg2='lightgreen', axes=11, camera=cam).close()



"""Delaunay 3D tetralization"""
from vedo import *
import numpy as np

settings.use_depth_peeling = True

pts = (np.random.rand(10000, 3)-0.5)*2

s = Sphere().alpha(0.1)
pin = s.inside_points(pts)
pin.subsample(0.05)  # impose min separation (5% of bounding box)
printc("# of points inside the sphere:", pin.npoints)

tmesh = delaunay3d(pin).shrink(0.95)

cmesh = tmesh.cut_with_plane(normal=(1,2,-1))
# cmesh.pipeline.show()  # to show the graph of operations

show([(s, pin, "Generate points in a Sphere"),
      (cmesh, __doc__),
     ], N=2, axes=1,
).close()



"""A polynomial fit of degree="""
from vedo import np, precision, Text3D, DashedLine, show
from vedo.pyplot import plot, fit, histogram
# np.random.seed(0)

n   = 25 # nr of data points to generate
deg = 3  # degree of the fitting polynomial

# Generate some noisy data points
x = np.linspace(0, 12, n)
y = (x-6)**3 /50 + 6              # the "truth" is a cubic function!
xerrs = np.linspace(0.4, 1.0, n)  # make last points less precise
yerrs = np.linspace(1.0, 0.4, n)  # make first points less precise
noise = np.random.randn(n)

# Plot the noisy points with their error bars
fig1 = plot(
    x, y+noise,
    ".k",
    title=__doc__+str(deg),
    xerrors=xerrs,
    yerrors=yerrs,
    aspect=4/5,
    xlim=(-3,15),
    ylim=(-3,15),
    padding=0,
)
fig1 += DashedLine(x, y, c='r')

# Fit points and evaluate, with a bootstrap and Monte-Carlo technique,
# the correct errors and error bands. Return a Line object:
pfit = fit(
    [x, y+noise],
    deg=deg,        # degree of the polynomial
    niter=500,      # nr. of MC iterations to compute error bands
    nstd=2,         # nr. of std deviations to display
    xerrors=xerrs,  # optional array of errors on x
    yerrors=yerrs,  # optional array of errors on y
    vrange=(-3,15), # specify the domain of fit
)

fig1 += [pfit, pfit.error_band, pfit.error_lines] # add these objects to Figure

# Add some text too
txt = "fit coefficients:\n " + precision(pfit.coefficients, 2) \
    + "\n:pm" + precision(pfit.coefficient_errors, 2) \
    + "\n:Chi^2_:nu  = " + precision(pfit.reduced_chi2, 3)
fig1 += Text3D(txt, s=0.42, font='VictorMono').pos(4,-2).c('k')

# Create a 2D histo to show the correlation of fit parameters
fig2 = histogram(
    pfit.monte_carlo_coefficients[:,0],
    pfit.monte_carlo_coefficients[:,1],
    title="parameters correlation",
    xtitle='coeff_0', ytitle='coeff_1',
    cmap='ocean_r',
    scalarbar=True,
)

show(fig1, fig2, N=2, sharecam=False, zoom='tight').close()



"""Extruding a polygon along the z-axis"""
from vedo import Star, show

# Create a yellow star and rotate it around the x-axis
star = Star().color('y').rotate_x(10)

# Extrude the star along the z-axis, with a shift of 1,
#  a rotation of 10 degrees, a decrease in radius of 0.2,
epol = star.extrude(zshift=1, rotation=10, dR=-0.2, cap=False, res=1)

# Set the back color of the extruded polygon to violet
epol.bc('violet').lighting("default")

# Show the extruded polygon, the script docstring, axes,
# then close the window
show(epol, __doc__, axes=1, viewup='z').close()



"""Double pendulum in 3D"""
# Original idea and solution using sympy from:
# https://www.youtube.com/watch?v=MtG9cueB548
from vedo import *

# Load the solution:
x1, y1, z1,  x2, y2, z2 = np.load(download(dataurl+'3Dpen.npy'))
p1, p2 = np.c_[x1,y1,z1], np.c_[x2,y2,z2]

ball1 = Sphere(c="green5", r=0.1, pos=p1[0])
ball2 = Sphere(c="blue5",  r=0.1, pos=p2[0])

ball1.add_shadow('z', -3)
ball2.add_shadow('z', -3)

ball1.add_trail(n=10)
ball2.add_trail(n=10)
ball1.trail.add_shadow('z', -3) # make trails project a shadow too
ball2.trail.add_shadow('z', -3)

rod1 = Line([0,0,0], ball1, lw=4)
rod2 = Line(ball1, ball2, lw=4)

axes = Axes(xrange=(-3,3), yrange=(-3,3), zrange=(-3,3))

# show the solution
plt = Plotter(interactive=False)
plt.show(ball1, ball2, rod1, rod2, axes, __doc__, viewup='z')

i = 0
for b1, b2 in zip(p1,p2):
    ball1.pos(b1)
    ball2.pos(b2)
    rod1.stretch([0,0,0], b1)
    rod2.stretch(b1, b2)
    ball1.update_shadows().update_trail()
    ball2.update_shadows().update_trail()
    ball1.trail.update_shadows()
    ball2.trail.update_shadows()
    plt.render()
    i+=1
    if i > 150:
        break

plt.interactive().close()



"""Share the same color map
across different meshes"""
from vedo import Mesh, show, dataurl


#####################################
man1 = Mesh(dataurl+"man.vtk")
scals = man1.points()[:, 2] * 5 + 27  # pick z coordinates [18->34]
man1.cmap("rainbow", scals, vmin=18, vmax=44)

#####################################
man2 = Mesh(dataurl+"man.vtk")
scals = man2.points()[:, 2] * 5 + 37  # pick z coordinates [28->44]
man2.cmap("rainbow", scals, vmin=18, vmax=44).add_scalarbar()

show([(man1, __doc__), man2], N=2, elevation=-40, axes=11).close()



"""Take 2 clouds of points, source and target, and morph
the plane using thin plate splines as a model.
The fitting minimizes the distance to a subset of the target cloud"""
from vedo import printc, Points, Grid, Arrows, Lines, Plotter
import scipy.optimize as opt
import numpy as np

np.random.seed(2)


class Morpher(Plotter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.source = None
        self.morphed_source = None
        self.target = None
        self.bound = 1
        self.sigma = 1  # stiffness of the mesh
        self.method = "SLSQP"  # 'SLSQP', 'L-BFGS-B', 'TNC' ...
        self.fitTolerance = 1e-6
        self.fitResult = None
        self.chi2 = 1.0e30
        self.npts = None
        self.ptsource = []
        self.pttarget = []

    def _func(self, pars):
        shift = np.array(np.split(pars, 2)).T  # recreate the shift vectors
        z = np.zeros((self.npts, 1))
        shift = np.append(shift, z, axis=1)  # make them 3d
        self.morphed_source = self.source.clone().warp(
            self.ptsource, self.ptsource + shift, sigma=self.sigma, mode="2d"
        )
        d = self.morphed_source.points() - self.target.points()
        chi2 = np.sum(np.multiply(d, d))  # /len(d)
        if chi2 < self.chi2:
            printc("new minimum ->", chi2)
            self.chi2 = chi2
        return chi2

    # ------------------------------------------------------- Fit
    def morph(self):
        print("\n..minimizing with " + self.method)
        self.morphed_source = self.source.clone()

        self.ptsource = self.source.points()[: self.npts]  # pick the first npts points
        self.pttarget = self.target.points()[: self.npts]

        delta = self.pttarget - self.ptsource
        x0 = delta[:, (0, 1)].T.ravel()  # initial guess, a flat list of x and y shifts
        bnds = [(-self.bound, self.bound)] * (2 * self.npts)
        res = opt.minimize(
            self._func, x0, bounds=bnds, method=self.method, tol=self.fitTolerance
        )
        self.fitResult = res
        # recalculate the last step:
        self._func(res["x"])

    # ------------------------------------------------------- Visualization
    def draw_shapes(self):
        sb = self.source.bounds()
        x1, x2, y1, y2, _, _ = sb
        maxb = max(x2 - x1, y2 - y1)
        grid0 = Grid(self.source.center_of_mass(), s=[maxb, maxb], res=[40, 40])
        T = self.morphed_source.transform
        grid1 = grid0.clone().apply_transform(T)  # warp the grid
        arrows = Arrows(self.ptsource, self.pttarget, alpha=0.5, s=3).c("k")
        lines = Lines(self.source, self.target).c("db")
        mlines = Lines(self.morphed_source, self.target).c("db")

        self.at(0).show(grid0, self.source, self.target, lines, arrows, __doc__)
        self.at(1).show(
            grid1,
            self.morphed_source,
            self.target,
            mlines,
            f"morphed source (green) vs target (red)\nNDF = {2*self.npts}",
        )


#################################
if __name__ == "__main__":

    # make up a source random cloud
    pts_s = np.random.randn(25, 2)
    pts_t = pts_s + np.sin(2 * pts_s) / 5  # and distort it

    mr = Morpher(N=2)
    mr.source = Points(pts_s, r=20, c="g", alpha=0.5)
    mr.target = Points(pts_t, r=10, c="r", alpha=1.0)

    mr.bound = 2  # limits the x and y shift
    # allow move only a subset of points (implicitly sets the NDF of the fit)
    mr.npts = 6
    mr.sigma = 1.0  # stiffness of the mesh (1=max stiffness)

    mr.morph()

    # now mr.msource contains the modified/morphed source.
    mr.draw_shapes()
    mr.interactive().close()



"""Scene interaction styles"""
from vedo import *

msg = Text2D(
    """TrackballCamera is the default mode\n(press q to proceed)""",
    c="k", bg="yellow7", s=1.2,
)

plt = Plotter(shape=(1,2))
plt.at(0).show(Cube(), msg).interactive()

msg.text("..lets change it to JoystickCamera").background("indigo7")
plt.at(1).show(Paraboloid(), mode="JoystickCamera").interactive()

msg.text("..lets change it again to MousePan").background("red6")
mode = interactor_modes.MousePan()
plt.user_mode(mode).interactive()

plt.close()



#!/usr/bin/env python3
"""An initial circle is stretched by means of a variable force into its final shape.
Colored lines are the trajectories of a few initial points."""
from dolfin import *
from mshr import *
import numpy as np
import vedo
from vedo.dolfin import plot

set_log_level(30)


class AllBoundaries(SubDomain):
    def inside(self, x, on_boundary):
        return on_boundary and x[0]<-10

def solve_problem(mesh, mfunc, force):
    V = VectorFunctionSpace(mesh, "CG", 1)
    u = TrialFunction(V)
    v = TestFunction(V)
    displacement = Function(V)

    bc = [DirichletBC(V, Constant((0,0)), mfunc, 1)]

    F = Constant(force)
    E = Constant(5000)
    nu = Constant(0.3)
    mu = E / (2.0*(1+nu))
    lmbda = E * nu / ((1.0+nu) * (1.0-2*nu))
    sigma = 2.0 * mu * sym(grad(u)) + lmbda * tr( sym(grad(u)) ) * Identity(2)
    solve(inner(sigma, grad(v)) * dx == inner(F, v) * dx, displacement, bc)
    displacement.set_allow_extrapolation(True)
    return displacement

def update(mesh, displacement):
    new_mesh = Mesh(mesh)
    ALE.move(new_mesh, displacement)
    return new_mesh

def remesh(mesh, res=50):
    if isinstance(mesh, vedo.Mesh):
        vmesh = mesh
    else:
        vmesh = vedo.Mesh([mesh.coordinates(), mesh.cells()])
    bpts = vmesh.computeNormals(cells=True).boundaries().join(reset=1) #extract boundary
    vz = vmesh.celldata["Normals"][0][2] # check z component of normals at first point
    bpts.generate_mesh(invert=vz<0).smooth().write('tmpmesh.xml') #vedo REMESHING + smoothing
    return Mesh("tmpmesh.xml")

#################################################################################
N = 40             # number of iterations of stretching
res = 15           # resolution of meshes
do_remesh = False  # grab the boundary and remesh the interior at each iteration
vedo.settings.use_parallel_projection = True  # avoid perspective parallax

circle = Circle(Point(0, 0), 50)
mesh = generate_mesh(circle, res)

meshes = [mesh]
displacements = []
for i in range(N):
    mfunc = MeshFunction('size_t', mesh, 1, mesh.domains())
    mfunc.set_all(0)
    allb = AllBoundaries()
    allb.mark(mfunc, 1)

    F = np.array([2, (i-N/2)/N]) # some variable force

    displacement = solve_problem(mesh, mfunc, F)
    new_mesh = update(mesh, displacement)

    if do_remesh:
        mesh = remesh(new_mesh)
    else:
        mesh = new_mesh
    meshes.append(mesh)
    displacements.append(displacement)
    # plot things:
    txt = vedo.Text2D(f"step{i}")
    arrow = vedo.Arrow2D([0,0], F*20).z(1)
    plot(mesh, arrow, txt, c='grey5', at=i, N=N, zoom=1.1) #PRESS q

dmesh_i = meshes[0]  # initial mesh
dmesh_f = meshes[-1] # final mesh

vmesh_i = vedo.Mesh([dmesh_i.coordinates(), dmesh_i.cells()], c='grey5').z(-1)
vmesh_f = vedo.Mesh([dmesh_f.coordinates(), dmesh_f.cells()], c='grey3').wireframe()

plt = vedo.Plotter()

# move a few points along the deformation of the circle
seeds = vedo.Circle(r=50, res=res).points()[:,(0,1)] # make points 2d with [:,(0,1)]
endpoints = []
for i, p in enumerate(seeds):
    line = [p]
    for u in displacements:
        p = p + u(p)
        line.append(p)
    plt += vedo.Line(line, c=i, lw=4).z(1)
    endpoints.append(p)

plt += [vmesh_i, vmesh_f, __doc__]
plt.show(axes=1)

# to invert everything and move the end points back in place, check out discussion:
#https://fenicsproject.discourse.group/t/precision-in-hyperelastic-model/6824/3





import wx
from vtk.wx.wxVTKRenderWindowInteractor import wxVTKRenderWindowInteractor
import vedo

#####################################################
# Every wx app needs an app
app = wx.App(False)

# create the top-level frame, sizer and wxVTKRWI
frame = wx.Frame(None, -1, "vedo with wxpython", size=(600,600))
widget = wxVTKRenderWindowInteractor(frame, -1)
sizer = wx.BoxSizer(wx.VERTICAL)
sizer.Add(widget, 1, wx.EXPAND)
frame.SetSizer(sizer)
frame.Layout()

# It would be more correct (API-wise) to call widget.Initialize() and
# widget.Start() here, but Initialize() calls RenderWindow.Render().
# That Render() call will get through before we can setup the
# RenderWindow() to render via the wxWidgets-created context; this
# causes flashing on some platforms and downright breaks things on
# other platforms.  Instead, we call widget.Enable().
widget.Enable(1)
widget.AddObserver("ExitEvent", lambda o,e,f=frame: f.Close())

##################################################### vedo example
def func(evt):
    print("Event dump:\n", evt)
    plt.azimuth(10) # rotate by one degree the camera

cone = vedo.shapes.Cone(c='green8')
axes = vedo.Axes(cone, c='white')
cube = vedo.shapes.Cube()

# Create 2 subwindows with a cone and a cube
plt = vedo.Plotter(N=2, bg='blue2', bg2='blue8', wx_widget=widget)
plt.add_callback("right mouse click", func)
plt.at(0).add([cone, axes, "right-click anywhere"]).reset_camera()
plt.at(1).add(cube).reset_camera()
# plt.show() # vedo.show() is now disabled in wx

#####################################################
# Show everything
frame.Show()
app.MainLoop()



# Plot a volume evolution in time
# Credits: https://github.com/edmontz
import numpy as np
from scipy.fftpack import fftn, fftshift
from vedo import Axes, Plotter, Volume, progressbar


def f(x, y, z, t):
    r = np.sqrt(x*x + y*y + z*z + 2*t*t) + 0.1
    return np.sin(9*np.pi * r)/r

n = 64
qn = 25
vol = np.zeros((n, n, n))
n1 = int(n/2)

plt = Plotter(bg="black", interactive=False)
axes = Axes(xrange=(0,n), yrange=(0,n), zrange=(0,n))
plt.show(axes, viewup='z')

for q in progressbar(range(qn), c='r'):
    t = 2 * q / qn - 1
    for k in range(n1):
        z = 2 * k / n1 - 1
        for j in range(n1):
            y = 2 * j / n1 - 1
            for i in range(n1):
                x = 2 * i / n1 - 1
                vol[i, j, k] = f(x, y, z, t)
    volf = fftn(vol)
    volf = fftshift(abs(volf))
    volf = np.log(12*volf/volf.max()+ 1) / 2.5

    vb = Volume(volf).mode(1).c("rainbow").alpha([0, 0.8, 1])
    vb.name = "MyVolume"
    plt.remove("MyVolume").add(vb).render()

plt.interactive().close()



"""The 3D-ruler axis style,
a flag pole and a goniometer"""
from vedo import *

settings.use_parallel_projection = True  # avoid parallax effects

mesh = Cone().c("steelblue").rotate_y(90).pos(1, 2, 3)

# add a flagpole-style comment
a, v = precision(mesh.area(), 4), precision(mesh.volume(), 4)
fp = mesh.flagpole(
    "S = πr^2 +πr√(h^2 +r^2 )\n  = " + a 
    + "~μm^2 \nV = πr^2 ·h/3\n  = " + v 
    + "~μm^3",
    s=0.1,
)

fp.color("r3").scale(0.7)

# measure the angle formed by 3 points
gon = Goniometer(
    [-0.5, 1, 2], [2.5, 2, 2], [-0.5, 3, 3], prefix=":alpha_c =~", lw=2, s=0.8
)

# show distance of 2 points
rul = Ruler(
    (-0.5, 2, 1.9),
    (2.5, 2, 2.9),
    prefix="L_x =",
    units="μm",
    axis_rotation=90,
    tick_angle=70,
)

# make 3d rulers along the bounding box (similar to set axes=7)
ax3 = RulerAxes(mesh, units="μm")

show(mesh, fp.follow_camera(), gon, rul, ax3, __doc__, bg2="lb", viewup="z").close()



"""Represent a volume as lego blocks (voxels).
Colors correspond to the volume's scalar.
Try also:
> vedo --lego data/embryo.tif"""
from vedo import *

vol = Volume(dataurl+'embryo.tif')

vol.crop(back=0.50) # crop 50% from neg. y

# show lego blocks whose value is between vmin and vmax
lego = vol.legosurface(vmin=20, vmax=None, boundary=False)
lego.cmap('seismic', vmin=0, vmax=127).add_scalarbar()

show(lego, __doc__, axes=1, viewup='z').close()



"""Press c while hovering to warp a Mesh onto another Mesh"""
from vedo import *


def on_keypress(event):
    if event.actor and event.keypress == "c":
        picked = event.picked3d
        idx = mesh.closest_point(picked, return_point_id=True)
        pt = points[idx]
        n = normals[idx]
        pt = pt + n / 5

        txt.orientation(n).pos(pt)

        tpts = txt.clone().subsample(0.05).points()
        kpts = [mesh.closest_point(tp) for tp in tpts]
        warped = txt.clone().warp(tpts, kpts, sigma=0.01, mode="2d")
        warped.c("purple5")

        lines = Lines(tpts, kpts, alpha=0.2)
        plt.remove("Text3D", "Lines").add(txt, warped, lines).render()


txt = Text3D("Text3D\n01-ABCD", s=0.1, justify="centered", c="red5")

mesh = ParametricShape("RandomHills").scale([1,1,0.5])
mesh.c("gray5").alpha(0.25)
points = mesh.points()
normals = mesh.normals()

plt = Plotter()
plt.add_callback("key press", on_keypress)
plt.show(mesh, txt, __doc__, axes=9, viewup="z").close()



"""Voronoi tessellation of a pointcloud on a grid"""
from vedo import dataurl, Points, Grid, voronoi, show

pts0 = Points(dataurl+'rios.xyz').color('k')
pts1 = pts0.clone().smooth_lloyd_2d()

grid = Grid([14500,61700], s=[22000,24000], res=[30,30]).ps(1)
allpts = pts1.points().tolist() + grid.points().tolist()

msh = voronoi(allpts, method='scipy')

msh.lw(0.1).wireframe(False).cmap('terrain_r', 'VoronoiID', on='cells')
centers = Points(msh.cell_centers(), c='k')

show(msh, pts0, __doc__, axes=dict(digits=3), zoom=1.3)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
from vedo import printc, Text2D, Text3D, show, Plotter
from vedo import fonts, fonts_path, settings
import numpy as np

################################################################################## 2D
inred = Text2D("°monospaced fonts are marked in red", c='r5', pos='bottom-center', font='VictorMono')
acts2d = [inred]

txt = 'The quick fox jumps over the lazy dog. 1234567890 αβγδεθλμνπστφψω'
for i, f in enumerate(fonts):
    bg = None
    if f in ['Calco', 'Glasgo', 'SmartCouric', 'VictorMono']:
        bg = 'red5'
    t = Text2D(f'{f}: {txt}', pos=(.015, 1-(i+3)*.06), font=f, s=1.3, c='k', bg=bg)
    acts2d.append(t)

acts2d.append(Text2D("List of built-in fonts", pos='top-center', bg='k', s=1.3))
plt0a = show(acts2d, bg2='cornsilk', size=(1300,800), interactive=False)

## online fonts:
acts2d = []
i = 0
for key, props in sorted(settings.font_parameters.items()):
    if props['islocal']:
        continue
    if key in ('Justino2', 'Justino3'):
        continue
    bg = None
    if props['mono']:
        bg = 'red5'
    t = Text2D(f'{key}: {txt}', pos=(.015, 1-(i+2)*.06), font=key, s=1.3, c='k', bg=bg)
    acts2d.append(t)
    i+=1
plt0b = show(acts2d,
             Text2D("Additional fonts (https://vedo.embl.es/fonts)", pos='top-center', bg='k', s=1.3),
             bg2='lb', size=(1300,900), pos=(1200,200), new=True)

################################################################################ printout
for font in fonts:
    printc(font + " - available characters:", " "*25, bold=1, invert=1)
    fontfile = os.path.join(fonts_path, font + '.npz')
    font_meshes = np.load(fontfile, allow_pickle=True)['font'][0]
    for k in font_meshes.keys():
        printc(k, end=' ')
    print()

printc('\n(use the above to copy&paste any char into your python script!)', italic=1)
printc('Symbols ~ ^ _ are reserved modifiers:', italic=1)
printc(' use ~ to add a short space, 1/4 of the default size,', italic=1)
printc(' use ^ and _ to start up/sub scripting, space terminates them.\n', italic=1)


################################################################################## 3D
# Symbols ~ ^ _ are reserved modifiers:
#  use ~ to add a short space, 1/4 of the default size,
#  use ^ and _ to start up/sub scripting, a space terminates them.
txt = """The quick fox jumps over the lazy dog.
Symbols: !@#$%&*()+=-{}[]:;|<>?/:euro1234567890
Units:  :delta=0.25E-03 ~μm, T_sea ~=~5.3~±0.7~:circC
LaTeX: :nabla:dotE=~4:pi~:rho, :nabla:timesE=~-1/c~~:partialB/:partialt
       ih~:partial/:partialt~:Psi = [-h^2 /2m:nabla^2  + V(r,t)]~:Psi(r,t)
       :DeltaE~=~h:nu, y = :Sigma_n ~A_n cos(:omega_n t+:delta_n ) sin(k_n x)
       :intx:dot~dx = :onehalf x:^2 + const.
       d^2 x^:mu  + :Gamma^:mu_:alpha:beta ~dx^:alpha ~dx^:beta  = 0
       -∇:^2u(x) = f(x) in Ω, u(x)~=~u_D (x) in :partial:Omega
Protect underscore \\\_ and \\\^ with a backslash.
"""

plt = Plotter(N=4, pos=(300,0), size=(1600,950))

cam = dict(pos=(3.99e+5, 8.51e+3, 6.47e+5),
           focal_point=(2.46e+5, 1.16e+5, -9.24e+3),
           viewup=(-0.0591, 0.983, 0.175),
           distance=6.82e+5,
           clipping_range=(5.26e+5, 8.92e+5))

for i,fnt in enumerate(["Kanopus", "Normografo", "Theemim", "VictorMono"]):
    t = Text3D(txt, font=fnt, italic=0).c('darkblue').scale(12300)
    plt.at(i)
    plt.show(t,
             Text2D("Font: "+fnt, font=fnt, bg='r'),
             axes=dict(xtitle='my units for L_x  (:mum)',
                       ytitle='my Y-axis with:na long description',
                       title_font=fnt,
                       label_font=fnt,
                       digits=2,
                      ),
             camera=cam,
             resetcam=not bool(i),
    )
plt.interactive().close()

plt0b.close()
plt0a.close()



"""Earthquakes of magnitude 2.5+ in the past 30 days
areas are proportional to energy release
[hover mouse to get more info]"""
import pandas
from vedo import *

num = 50  # nr of earthquakes to be visualized to define a time window
path = download("https://earthquake.usgs.gov/earthquakes/feed/v1.0/summary/2.5_month.csv")
usecols = ['time','place','latitude','longitude','depth','mag']
data = pandas.read_csv(path, usecols=usecols)[usecols][::-1].reset_index(drop=True) # reverse list

pic = Picture(dataurl + "images/eo_base_2020_clean_3600x1800.png")
pic.pickable(False).level(185).window(120)  # add some contrast to the original image
scale = [pic.shape[0]/2, pic.shape[1]/2, 1]
comment = Text2D(__doc__, bg='green9', alpha=0.7, font='Ubuntu')

centers = []
for i, d in progressbar(data.iterrows()):
    M = d['mag']                                       # earthquake estimated magnitude
    E = np.sqrt(np.exp(5.24+1.44*M) * scale[0])/10000  # empirical formula for sqrt(energy_release(M))
    rgb = color_map(E, name='Reds', vmin=0, vmax=7)    # map energy to color
    lat = np.deg2rad(d['latitude'])
    lon = np.deg2rad(d['longitude'])
    ce = GeoCircle(lat, lon, E/50).scale(scale).z(num/M)
    ce.c(rgb).alpha(0.75).use_bounds(False)
    ce.time = i
    ce.info = '\n'.join(str(d).split('\n')[:-1])       # remove of the last line in string d
    if i < len(data) - num: 
        ce.off()                                       # switch off older ones: make circles invisible
    centers.append(ce)


def sliderfunc(widget, event):
    val = widget.value                                 # get the slider current value
    widget.title = f"{data['time'][int(val)][:10]}"
    for ce in centers:
        isinside = abs(val-ce.time) < num              # switch on if inside of time window
        ce.on() if isinside else ce.off()

plt = Plotter(size=(2200,1100), title="Earthquake Browser").parallel_projection()
plt.add_slider(sliderfunc, 0, len(centers)-1, value=len(centers)-1, show_value=False, title="today")
plt.add_hover_legend(use_info=True, alpha=1, c='white', bg='red2', s=1)
plt.show(pic, centers, comment, zoom="tightest", mode='image').close()



# A plot(mode="bars") example. Useful to plot categories.
from vedo import precision, Text3D, color_map, settings
from vedo.pyplot import plot

settings.default_font = "Meson"

counts  = [1946, 8993, 3042, 1190, 1477,    0,    0]
percent = [11.68909178, 54.01850072, 18.27246516,  7.14800577,  8.87193657, 0, 0]
labels  = ['<100', '100-250', '250-500', '500-750', '750-1000', '1000-2000', '>2000']
colors  = color_map(range(len(counts)), "hot")

# plot() will return a PlotBars object
fig = plot(
    [counts, labels, colors],
    mode="bars",
    ylim=(0,10_000),
    aspect=16/9,
    title='Clusters in lux range',
    axes=dict(
        xlabel_rotation=30,
        xlabel_size=0.02,
    ),
)

for i in range(len(percent)):
    val = precision(percent[i], 3)+'%'
    txt = Text3D(val, pos=(fig.centers[i], counts[i]), justify="bottom-center", c="blue2")
    fig += txt.scale(200).shift(0,170,0)

fig.show(size=(1000,750), zoom='tight').close()



"""Euclidean Distance Transform using Saito algorithm.
The distance map produced contains the square
of the Euclidean distance values"""
from vedo import *

e = Volume(dataurl+'embryo.tif')

edt = e.euclidean_distance()

show([(e,__doc__), edt], N=2, viewup='z', axes=1, zoom=1.5).close()




from vedo import dataurl, Plotter, Volume

plt = Plotter(N=2)

# Load a mesh and show it
vol = Volume(dataurl+"embryo.tif")
m0 = vol.isosurface(flying_edges=False).normalize().lw(1).c("violet")

# Smooth the mesh
m1 = m0.clone().smooth(niter=20).color("lg")

plt.at(0).show(m0, "Original Mesh:")
plt.background('light blue')  # set first renderer color

plt.at(1).show(
    "Mesh polygons are smoothed:", m1,
    viewup='z', zoom=1.5)
plt.interactive().close()



"""Manipulate a box to cut a mesh interactively.
Use mouse buttons to zoom and pan.
Press r to reset the cutting box
Press i to toggle the cutting box on/off
Press u to invert the selection"""
from vedo import *

# settings.enable_default_keyboard_callbacks = False
# settings.enable_default_mouse_callbacks = False

msh = Mesh(dataurl+'mouse_brain.stl').backcolor("purple8")

plt = Plotter(bg='blackboard', interactive=False)
plt.show(msh, __doc__, viewup='z')

cutter = PlaneCutter(msh)
# cutter = BoxCutter(msh)
# cutter = SphereCutter(msh)

plt.add(cutter)
plt.interactive()

plt.remove(cutter)
plt.interactive()

plt.close()



"""Draw the isolines and isobands of
a scalar field H (height) on a surface"""
from vedo import *

mesh0 = ParametricShape('RandomHills')

# ParametricShapes already have a scalar associated to points
printc('Mesh point arrays:', mesh0.pointdata.keys())
# so assign it a colormap:
mesh0.cmap('terrain')

isol = mesh0.isolines(n=10).color('w')
isob = mesh0.isobands(n=5).add_scalarbar("H=Elevation")

# make a copy and interpolate the Scalars from points to cells
mesh1 = mesh0.clone(deep=False).map_points_to_cells()
printc('Mesh cell arrays :', mesh1.celldata.keys())

gvecs = mesh1.gradient(on='cells')
cc = mesh1.cell_centers()
ars = Arrows(cc, cc + gvecs*0.01, c='bone_r').lighting('off')
ars.add_scalarbar3d(title='|:nablaH|:dot0.01 [arb.units]')

# colormap the gradient magnitude directly on the mesh
mesh2 = mesh1.clone(deep=False).lw(0.1).cmap('jet', mag(gvecs), on='cells')
mesh2.add_scalarbar3d(title='|:nablaH| [arb.units]')

plt = Plotter(N=4, size=(1200,900), axes=11)
plt.at(0).show(mesh0, isol, __doc__)
plt.at(1).show(isob)
plt.at(2).show(mesh1, isol, ars, "Arrows=:nablaH")
plt.at(3).show(mesh2, "Color=|:nablaH|")
plt.interactive().close()



"""Fit a plane to regions of a surface defined by
N points that are closest to a given point of the surface."""
from vedo import *


apple = Mesh(dataurl+"apple.ply").subdivide().add_gaussian_noise(0.5)

plt = Plotter()
plt += apple.alpha(0.1)

variances = []
for i, p in enumerate(apple.points()):
    pts = apple.closest_point(p, n=12) # find the N closest points to p
    plane = fit_plane(pts)             # find the fitting plane
    variances.append(plane.variance)
    if i % 200: continue
    plt += plane
    plt += Points(pts)
    plt += Arrow(plane.center, plane.center+plane.normal/5)

plt += __doc__ + "\nNr. of fits performed: "+str(len(variances))
plt.show().close()




import wx
import vedo
from vtk.wx.wxVTKRenderWindowInteractor import wxVTKRenderWindowInteractor

##################################################### wx app
app = wx.App(False)
frame = wx.Frame(None, -1, "vedo with wxpython", size=(800,800))
widget = wxVTKRenderWindowInteractor(frame, -1)
sizer = wx.BoxSizer(wx.VERTICAL)
sizer.Add(widget, 1, wx.EXPAND)
frame.SetSizer(sizer)
frame.Layout()
widget.Enable(1)
widget.AddObserver("ExitEvent", lambda o,e,f=frame: f.Close())

##################################################### vedo
def func(event):
    mesh = event.actor
    if not mesh: return

    ptid = mesh.closest_point(event.picked3d, return_point_id=True)
    txt = f"Probed point:\n{vedo.utils.precision(event.picked3d, 3)}\n" \
          f"value = {vedo.utils.precision(arr[ptid], 2)}"

    vpt = vedo.shapes.Sphere(mesh.points(ptid), r=0.01, c='orange2').pickable(False)
    vig = vpt.flagpole(txt, s=.05, offset=(0.5,0.5), font="VictorMono").follow_camera()

    msg.text(txt)               # update the 2d text message
    plt.remove(plt.actors[-2:]).add([vpt, vig]) # remove last 2 objects, add the new ones
    widget.Render()             # need to manually call Render

msg = vedo.Text2D(pos='bottom-left', font="VictorMono")
msh = vedo.shapes.ParametricShape("RandomHills").cmap('terrain')
axs = vedo.Axes(msh)
arr = msh.pointdata["Scalars"]

plt = vedo.Plotter(bg='moccasin', bg2='blue9', wx_widget=widget)
plt.add([msh, axs, msg]).reset_camera()
plt.actors += [None,None,None]  # place holder for sphere, flagpole, text2d
plt.add_callback('MouseMove', func)

#####################################################
# Show everything
frame.Show()
app.MainLoop()



"""Interactively cut a mesh by drawing free-hand a spline in space"""
# The tool can also be invoked from command line e.g.: > vedo --edit mesh.ply
import vedo
from vedo.applications import FreeHandCutPlotter

#### This class is a simplified version of the above, shown here as an example: #######
#
# class FreeHandCutPlotter(vedo.Plotter):
#     def __init__(self, mesh):
#         vedo.Plotter.__init__(self)
#         self.mesh = mesh
#         self.drawmode = False
#         self.cpoints = []
#         self.points = None
#         self.spline = None
#         self.msg  = "Right-click and move to draw line\n"
#         self.msg += "Second right-click to stop drawing\n"
#         self.msg += "Press z to cut mesh"
#         self.txt2d = vedo.Text2D(self.msg, pos='top-left', font="Bongas")
#         self.txt2d.c("white").background("green4", alpha=1)
#         self.add_callback('KeyPress', self.onKeyPress)
#         self.add_callback('RightButton', self.onRightClick)
#         self.add_callback('MouseMove', self.onMouseMove)

#     def onRightClick(self, evt):
#         self.drawmode = not self.drawmode  # toggle mode

#     def onMouseMove(self, evt):
#         if self.drawmode:
#             self.remove([self.points, self.spline])
#             # make this 2d-screen point 3d:
#             cpt = self.compute_world_coordinate(evt.picked2d)
#             self.cpoints.append(cpt)
#             self.points = vedo.Points(self.cpoints, r=8).c('black')
#             if len(self.cpoints) > 2:
#                 self.spline = vedo.Line(self.cpoints, closed=True).lw(5).c('red5')
#                 self.add([self.points, self.spline]).render()

#     def onKeyPress(self, evt):
#         if evt.keypress == 'z' and self.spline:  # cut mesh with a ribbon-like surf.
#             vedo.printc("Cutting the mesh please wait..", invert=True)
#             tol = self.mesh.diagonal_size()/2            # size of ribbon
#             pts = self.spline.points()
#             n = vedo.fit_plane(pts, signed=True).normal  # compute normal vector
#             rib = vedo.Ribbon(pts - tol*n, pts + tol*n, closed=True)
#             self.mesh.cut_with_mesh(rib)
#             self.remove([self.spline, self.points]).render()
#             self.cpoints, self.points, self.spline = [], None, None

#     def start(self, **kwargs):
#         return self.show(self.txt2d, self.mesh, **kwargs)
#
######################################################################################
vedo.settings.use_parallel_projection = True  # to avoid perspective artifacts

msh = vedo.Volume(vedo.dataurl+'embryo.tif').isosurface().color('gold', 0.25) # Mesh

plt = FreeHandCutPlotter(msh).add_hover_legend()
#plt.init(some_list_of_initial_pts) #optional!
plt.start(axes=1, bg2='lightblue').close()



"""
This demo solves the Stokes equations, using quadratic elements for
the velocity and first degree elements for the pressure (Taylor-Hood elements).
"""
# Credits:
# https://github.com/pf4d/fenics_scripts/blob/master/cbc_block/stokes.py
from dolfin import *
import numpy as np
from vedo.dolfin import plot
from vedo import Latex, dataurl, download

# Load mesh and subdomains
fpath = download(dataurl + "dolfin_fine.xml")
mesh = Mesh(fpath)

fpath = download(dataurl + "dolfin_fine_subdomains.xml.gz")
sub_domains = MeshFunction("size_t", mesh, fpath)

# Define function spaces
P2 = VectorElement("Lagrange", mesh.ufl_cell(), 2)
P1 = FiniteElement("Lagrange", mesh.ufl_cell(), 1)
TH = P2 * P1
W = FunctionSpace(mesh, TH)

# No-slip boundary condition for velocity
noslip = Constant((0, 0))
bc0 = DirichletBC(W.sub(0), noslip, sub_domains, 0)

# Inflow boundary condition for velocity
inflow = Expression(("-sin(x[1]*pi)", "0.0"), degree=2)
bc1 = DirichletBC(W.sub(0), inflow, sub_domains, 1)
bcs = [bc0, bc1]

# Define variational problem
(u, p) = TrialFunctions(W)
(v, q) = TestFunctions(W)
f = Constant((0, 0))
a = (inner(grad(u), grad(v)) - div(v) * p + q * div(u)) * dx
L = inner(f, v) * dx
w = Function(W)

solve(a == L, w, bcs)

# Split the mixed solution using a shallow copy
(u, p) = w.split()

##################################################################### vedo
f = r"-\nabla \cdot(\nabla u+p I)=f ~\mathrm{in}~\Omega"
formula = Latex(f, pos=(0.55, 0.45, -0.05), s=0.1)

plot(
    u,
    formula,
    at=0,
    N=2,
    mode="mesh and arrows",
    scale=0.03,
    wireframe=True,
    scalarbar=False,
    style=1,
)
plot(p, at=1, text="pressure", cmap="rainbow", interactive=True).close()


##################################################################### streamlines
# A list of seed points (can be automatic: just comment out 'probes')
ally = np.linspace(0, 1, num=100)
probes = np.c_[np.ones_like(ally), ally, np.zeros_like(ally)]

plot(
    u,
    mode="mesh with streamlines",
    streamlines={
        "tol": 0.02,  # control density of streams
        "lw": 2,  # line width
        "direction": "forward",  # direction of integration
        "maxPropagation": 2.2,  # max length of propagation
        "probes": probes,  # custom list of point in space as seeds
    },
    c="white",  # mesh color
    alpha=0.3,  # mesh alpha
    lw=0,  # mesh line width
    wireframe=True,  # show as wireframe
    bg="blackboard",  # background color
    new=True,  # new window
    pos=(200, 200),  # window position on screen
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os

__docformat__ = "google"


class Settings:
    """
    General settings to modify the global behavior

    Usage Example:
        ```python
        from vedo import settings, Cube
        settings.use_parallel_projection = True
        Cube().color('g').show().close()
        ```

    List of available properties:

    ```python
    # Set a default for the font to be used for axes, comments etc.
    default_font = 'Normografo' # check font options in shapes.Text

    # Palette number when using an integer to choose a color
    palette = 0

    screenshot_transparent_background = False
    screeshot_large_image = False # Sometimes setting this to True gives better results

    # [DISABLED] Allow to continuously interact with scene during interactive() execution
    allow_interaction = True

    # Enable tracking pipeline functionality: 
    #  allows to show a graph with the pipeline of action which let to a final object
    #  this is achieved by calling "myobj.pipeline.show()" (a new window will pop up)
    self.enable_pipeline = True

    # Set up default mouse and keyboard functionalities
    enable_default_mouse_callbacks = True
    enable_default_keyboard_callbacks = True

    # If False, when multiple renderers are present do not render each one for separate
    #  but do it just once at the end (when interactive() is called)
    immediate_rendering = True

    # Show a gray frame margin in multirendering windows
    renderer_frame_color = None
    renderer_frame_alpha = 0.5
    renderer_frame_width = 0.5
    renderer_frame_padding = 0.0001

    # In multirendering mode set the position of the horizontal of vertical splitting [0,1]
    window_splitting_position = None

    # Enable / disable color printing by printc()
    enable_print_color = True

    # Wrap lines in tubes
    render_lines_as_tubes = False

    # Smoothing options
    point_smoothing = False
    line_smoothing = False
    polygon_smoothing = False

    # Remove hidden lines when in wireframe mode
    hidden_line_removal = False

    # Turn on/off the automatic repositioning of lights as the camera moves.
    light_follows_camera = False
    two_sided_lighting = True

    # Turn on/off rendering of translucent material with depth peeling technique.
    use_depth_peeling = False
    alpha_bit_planes  = True   # options only active if useDepthPeeling=True
    multi_samples     = 8      # force to not pick a framebuffer with a multisample buffer
    max_number_of_peels= 4     # maximum number of rendering passes
    occlusion_ratio   = 0.0    # occlusion ratio, 0 = exact image.

    # Turn on/off nvidia FXAA post-process anti-aliasing, if supported.
    use_fxaa = False           # either True or False

    # By default, the depth buffer is reset for each renderer.
    #  If True, use the existing depth buffer
    preserve_depth_buffer = False

    # Use a polygon/edges offset to possibly resolve conflicts in rendering
    use_polygon_offset    = False
    polygon_offset_factor = 0.1
    polygon_offset_units  = 0.1

    # Interpolate scalars to render them smoothly
    interpolate_scalars_before_mapping = True

    # Set parallel projection On or Off (place camera to infinity, no perspective effects)
    use_parallel_projection = False

    # Set orientation type when reading TIFF files (volumes):
    # TOPLEFT  1 (row 0 top, col 0 lhs)    TOPRIGHT 2 (row 0 top, col 0 rhs)
    # BOTRIGHT 3 (row 0 bottom, col 0 rhs) BOTLEFT  4 (row 0 bottom, col 0 lhs)
    # LEFTTOP  5 (row 0 lhs, col 0 top)    RIGHTTOP 6 (row 0 rhs, col 0 top)
    # RIGHTBOT 7 (row 0 rhs, col 0 bottom) LEFTBOT  8 (row 0 lhs, col 0 bottom)
    tiff_orientation_type = 1

    # Annotated cube axis type nr. 5 options:
    annotated_cube_color      = (0.75, 0.75, 0.75)
    annotated_cube_text_color = None # use default, otherwise specify a single color
    annotated_cube_text_scale = 0.2
    annotated_cube_texts      = ["right","left ", "front","back ", " top ", "bttom"]

    # Automatically close the Plotter instance after show() in jupyter sessions
    # setting it to False will keep the current Plotter instance active
    backend_autoclose = True

    # k3d settings for jupyter notebooks
    k3d_menu_visibility = True
    k3d_plot_height = 512
    k3d_antialias   = True
    k3d_lighting    = 1.5
    k3d_camera_autofit = True
    k3d_grid_autofit= True
    k3d_axes_color  = "gray4"
    k3d_axes_helper = 1.0     # size of the small triad of axes on the bottom right
    k3d_point_shader= "mesh"  # others are '3d', '3dSpecular', 'dot', 'flat'
    k3d_line_shader = "thick" # others are 'flat', 'mesh'
    ```
    """

    # Restrict the attributes so accidental typos will generate an AttributeError exception
    __slots__ = [
        "_level",
        "default_font",
        "default_backend",
        "palette",
        "remember_last_figure_format",
        "screenshot_transparent_background",
        "screeshot_large_image",
        "allow_interaction",
        "hack_call_screen_size",
        "enable_default_mouse_callbacks",
        "enable_default_keyboard_callbacks",
        "enable_pipeline",
        "immediate_rendering",
        "renderer_frame_color",
        "renderer_frame_alpha",
        "renderer_frame_width",
        "renderer_frame_padding",
        "render_lines_as_tubes",
        "hidden_line_removal",
        "point_smoothing",
        "line_smoothing",
        "polygon_smoothing",
        "visible_grid_edges",
        "light_follows_camera",
        "two_sided_lighting",
        "use_depth_peeling",
        "multi_samples",
        "alpha_bit_planes",
        "max_number_of_peels",
        "occlusion_ratio",
        "use_fxaa",
        "preserve_depth_buffer",
        "use_polygon_offset",
        "polygon_offset_factor",
        "polygon_offset_units",
        "interpolate_scalars_before_mapping",
        "use_parallel_projection",
        "window_splitting_position",
        "tiff_orientation_type",
        "annotated_cube_color",
        "annotated_cube_text_color",
        "annotated_cube_text_scale",
        "annotated_cube_texts",
        "enable_print_color",
        "backend_autoclose",
        "k3d_menu_visibility",
        "k3d_plot_height",
        "k3d_antialias",
        "k3d_lighting",
        "k3d_camera_autofit",
        "k3d_grid_autofit",
        "k3d_axes_color",
        "k3d_axes_helper",
        "k3d_point_shader",
        "k3d_line_shader",
        "font_parameters",
    ]

    def __init__(self, level=0):

        self._level = level

        # Default font
        self.default_font = "Normografo"

        # Default backend engine in jupyter notebooks
        self.default_backend = "vtk"

        # enable tracking pipeline functionality
        self.enable_pipeline = True

        if any(["SPYDER" in name for name in os.environ]):
            self.default_backend = "vtk"
        else:
            try:
                get_ipython()
                self.default_backend = "2d"
            except NameError:
                pass

        # Palette number when using an integer to choose a color
        self.palette = 0

        self.remember_last_figure_format = False

        self.screenshot_transparent_background = False
        self.screeshot_large_image = False

        # [DISABLED] Allow to continuously interact with scene during interactor.Start() execution
        self.allow_interaction = True

        # BUG in vtk9.0 (if true close works but sometimes vtk crashes, if false doesnt crash but cannot close)
        # see plotter.py line 555
        self.hack_call_screen_size = True

        # Set up default mouse and keyboard functionalities
        self.enable_default_mouse_callbacks = True
        self.enable_default_keyboard_callbacks = True

        # When multiple renderers are present do not render each one for separate.
        # but do it just once at the end (when interactive() is called)
        self.immediate_rendering = True

        # Show a gray frame margin in multirendering windows
        self.renderer_frame_color = None
        self.renderer_frame_alpha = 0.5
        self.renderer_frame_width = 0.5
        self.renderer_frame_padding = 0.0001

        # Wrap lines in tubes
        self.render_lines_as_tubes = False

        # Remove hidden lines when in wireframe mode
        self.hidden_line_removal = False

        # Smoothing options
        self.point_smoothing = False
        self.line_smoothing = False
        self.polygon_smoothing = False

        # For Structured and RectilinearGrid: show internal edges not only outline
        self.visible_grid_edges = False

        # Turn on/off the automatic repositioning of lights as the camera moves.
        self.light_follows_camera = False
        self.two_sided_lighting = True

        # Turn on/off rendering of translucent material with depth peeling technique.
        self.use_depth_peeling = False
        self.multi_samples = 8
        self.alpha_bit_planes = 1
        self.max_number_of_peels = 4
        self.occlusion_ratio = 0.1

        # Turn on/off nvidia FXAA anti-aliasing, if supported.
        self.use_fxaa = False  # either True or False

        # By default, the depth buffer is reset for each renderer. If true, use the existing depth buffer
        self.preserve_depth_buffer = False

        # Use a polygon/edges offset to possibly resolve conflicts in rendering
        self.use_polygon_offset = True
        self.polygon_offset_factor = 0.1
        self.polygon_offset_units  = 0.1

        # Interpolate scalars to render them smoothly
        self.interpolate_scalars_before_mapping = True

        # Set parallel projection On or Off (place camera to infinity, no perspective effects)
        self.use_parallel_projection = False

        # In multirendering mode set the position of the horizontal of vertical splitting [0,1]
        self.window_splitting_position = None

        # Set orientation type when reading TIFF files (volumes):
        # TOPLEFT  1 (row 0 top, col 0 lhs)    TOPRIGHT 2 (row 0 top, col 0 rhs)
        # BOTRIGHT 3 (row 0 bottom, col 0 rhs) BOTLEFT  4 (row 0 bottom, col 0 lhs)
        # LEFTTOP  5 (row 0 lhs, col 0 top)    RIGHTTOP 6 (row 0 rhs, col 0 top)
        # RIGHTBOT 7 (row 0 rhs, col 0 bottom) LEFTBOT  8 (row 0 lhs, col 0 bottom)
        self.tiff_orientation_type = 1

        # AnnotatedCube axis (type 5) customization:
        self.annotated_cube_color = (0.75, 0.75, 0.75)
        self.annotated_cube_text_color = None  # use default, otherwise specify a single color
        self.annotated_cube_text_scale = 0.2
        self.annotated_cube_texts = ["right", "left ", "front", "back ", " top ", "bttom"]

        # enable / disable color printing
        self.enable_print_color = True

        ####################################################################################
        # Automatically close the Plotter instance after show() in jupyter sessions,
        #  setting it to False will keep the current Plotter instance active
        self.backend_autoclose = True

        # k3d settings for jupyter notebooks
        self.k3d_menu_visibility = True
        self.k3d_plot_height = 512
        self.k3d_antialias  = True
        self.k3d_lighting   = 1.5
        self.k3d_camera_autofit = True
        self.k3d_grid_autofit= True
        self.k3d_axes_color  = "k4"
        self.k3d_axes_helper = 1.0     # size of the small triad of axes on the bottom right
        self.k3d_point_shader= "mesh"  # others are '3d', '3dSpecular', 'dot', 'flat'
        self.k3d_line_shader = "thick" # others are 'flat', 'mesh'

        ####################################################################################
        ####################################################################################
        # mono       # means that all letters occupy the same space slot horizontally
        # hspacing   # an horizontal stretching factor (affects both letters and words)
        # lspacing   # horizontal spacing inbetween letters (not words)
        # islocal    # is locally stored in /fonts, otherwise it's on vedo.embl.es/fonts
        #
        self.font_parameters = dict(
            Normografo=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=True,
            ),
            Bongas=dict(
                mono=False,
                fscale=0.875,
                hspacing=0.52,
                lspacing=0.25,
                dotsep="·",
                islocal=True,
            ),
            Calco=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=True,
            ),
            Comae=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=True,
            ),
            ComicMono=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="x",
                islocal=False,
            ),
            Edo=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~x ",
                islocal=False,
            ),
            FiraMonoMedium=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            FiraMonoBold=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Glasgo=dict(
                mono=True,
                fscale=0.75,
                lspacing=0.1,
                hspacing=1,
                dotsep="~×",
                islocal=True,
            ),
            Kanopus=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.15,
                hspacing=0.75,
                dotsep="~×",
                islocal=True,
            ),
            LogoType=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Quikhand=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.6,
                lspacing=0.15,
                dotsep="~~×~",
                islocal=True,
            ),
            SmartCouric=dict(
                mono=True,
                fscale=0.8,
                hspacing=1.05,
                lspacing=0.1,
                dotsep="×",
                islocal=True,
            ),
            Spears=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.5,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Theemim=dict(
                mono=False,
                fscale=0.825,
                hspacing=0.52,
                lspacing=0.3,
                dotsep="~~×",
                islocal=True,
            ),
            VictorMono=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=True,
            ),
            Justino1=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Justino2=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Justino3=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="×",
                islocal=False,
            ),
            Calibri=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Capsmall=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.75,
                lspacing=0.15,
                dotsep="~×",
                islocal=False,
            ),
            Cartoons123=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.75,
                lspacing=0.15,
                dotsep="x",
                islocal=False,
            ),
            Vega=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.75,
                lspacing=0.15,
                dotsep="×",
                islocal=False,
            ),
            Meson=dict(
                mono=False,
                fscale=0.8,
                hspacing=0.9,
                lspacing=0.225,
                dotsep="~×",
                islocal=False,
            ),
            Komika=dict(
                mono=False,
                fscale=0.7,
                hspacing=0.75,
                lspacing=0.225,
                dotsep="~×",
                islocal=False,
            ),
            Vogue=dict(
                mono=False,
                fscale=0.7,
                hspacing=0.75,
                lspacing=0.225,
                dotsep="~x",
                islocal=False,
            ),
            Brachium=dict(
                mono=True,
                fscale=0.8,
                hspacing=1,
                lspacing=0.1,
                dotsep="x",
                islocal=False,
            ),
            Dalim=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Miro=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Ubuntu=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Mizar=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=0.75,
                dotsep="~×",
                islocal=False,
            ),
            LiberationSans=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            DejavuSansMono=dict(
                mono=True,
                fscale=0.725,
                hspacing=1,
                lspacing=0.1,
                dotsep="~×",
                islocal=False,
            ),
            SunflowerHighway=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Swansea=dict(
                mono=False,
                fscale=0.75,
                lspacing=0.2,
                hspacing=1,
                dotsep="~×",
                islocal=False,
            ),
            Housekeeper=dict(  # support chinese glyphs
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~×",
                islocal=False,
            ),
            Wananti=dict(  # support chinese glyphs
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~x",
                islocal=False,
            ),
            AnimeAce=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~x",
                islocal=False,
            ),
            AnimeAceBold=dict(
                mono=False,
                fscale=0.75,
                hspacing=1,
                lspacing=0.2,
                dotsep="~x",
                islocal=False,
            ),
        )

    ####################################################################################
    def reset(self):
        """Reset all settings to their default status."""
        self.__init__()

    def print(self):
        """Print function."""
        print(" " + "-" * 80)
        s = Settings.__doc__.replace("   ", "")
        s = s.replace(".. code-block:: python\n", "")
        try:
            from pygments import highlight
            from pygments.lexers import Python3Lexer
            from pygments.formatters import Terminal256Formatter

            s = highlight(s, Python3Lexer(), Terminal256Formatter(style="zenburn"))
            print(s, end="")

        except ModuleNotFoundError:
            print("\x1b[33;1m" + s + "\x1b[0m")

    def __getitem__(self, key):
        """Make the class work like a dictionary too"""
        return getattr(self, key)

    def __setitem__(self, key, value):
        """Make the class work like a dictionary too"""
        setattr(self, key, value)


    def init_colab(self, enable_k3d=True):
        """
        Initialize colab environment
        """
        print("setting up colab environment (can take a minute) ...", end='')

        res = os.system('which Xvfb')
        if res:
            os.system('apt-get install xvfb')

        os.system('pip install pyvirtualdisplay')

        from pyvirtualdisplay import Display
        Display(visible=0).start()

        if enable_k3d:
            os.system('pip install k3d')

        from google.colab import output
        output.enable_custom_widget_manager()

        if enable_k3d:
            import k3d
            try:
                print("installing k3d...", end='')
                os.system("jupyter nbextension install --py --user k3d")
                os.system("jupyter nbextension enable  --py --user k3d")
                k3d.switch_to_text_protocol()
                self.default_backend = 'k3d'
                self.backend_autoclose = False
            except:
                print("(FAILED) ... ", end='')

        print(" setup completed.")


    def start_xvfb(self):
        """
        Start xvfb.

        Xvfb or X virtual framebuffer is a display server implementing
        the X11 display server protocol. In contrast to other display servers,
        Xvfb performs all graphical operations in virtual memory
        without showing any screen output.
        """
        print("starting xvfb (can take a minute) ...", end='')
        res = os.system('which Xvfb')
        if res:
            os.system('apt-get install xvfb')
        os.system('set -x')
        os.system('export DISPLAY=:99.0')
        os.system('Xvfb :99 -screen 0 1024x768x24 > /dev/null 2>&1 &')
        os.system('sleep 3')
        os.system('set +x')
        os.system('exec "$@"')
        print(" xvfb started.")

